from abc import abstractmethod
import numpy as np
import pandas as pd
import warnings as w
from scipy.optimize import root_scalar
from scipy.optimize import root
from copy import deepcopy

from VESIcal import activity_models
from VESIcal import core
from VESIcal import fugacity_models


class Model(object):
    """The model object implements a volatile solubility model. It is composed
    of the methods needed to evaluate
    :func:`VESIcal.calculate_dissolved_volatiles`,
    :func:`VESIcal.calculate_equilibrium_fluid_comp`, and
    :func:`calculate_saturation_pressure`. The fugacity and activity models for
    the volatiles species must be specified, defaulting to ideal.
    """

    def __init__(self):
        self.set_volatile_species(None)
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([])
        self.set_solubility_dependence(False)

    def set_volatile_species(self, volatile_species):
        if type(volatile_species) == str:
            volatile_species = [volatile_species]
        elif type(volatile_species) != list:
            raise core.InputError("volatile_species must be a str or list.")
        self.volatile_species = volatile_species

    def set_fugacity_model(self, fugacity_model):
        self.fugacity_model = fugacity_model

    def set_activity_model(self, activity_model):
        self.activity_model = activity_model

    def set_calibration_ranges(self, calibration_ranges):
        self.calibration_ranges = calibration_ranges

    def set_solubility_dependence(self, solubility_dependence):
        self.solubility_dependence = solubility_dependence

    def get_calibration_values(self, variable_names):
        """ Returns the values stored as the calibration range for the given
        variable(s). However, for checks where there is a single value- i.e.
        cr_GreaterThan or crf_LessThan, the logical operation will remain a
        mystery until someone figures out an elegant way of communicating it.

        Parameters
        ----------
        variable_names     str or list
            The name(s) of the variables you want the calibration ranges for.

        Returns
        -------
        list
            A list of values or tuples for the calibration ranges in the order
            given.
        """

        # Check if the variable name is passed as a string, and if so put it in
        # a list:
        if type(variable_names) == str:
            variable_names = [variable_names]

        calibration_values = []

        for var in variable_names:
            found_var = False
            for cr in self.calibration_ranges:
                if found_var is False:
                    if cr.parameter_name == var:
                        found_var = True
                        calibration_values.append(cr.value)
            if found_var is False:
                calibration_values.append(np.nan)

        return calibration_values

    @abstractmethod
    def calculate_dissolved_volatiles(self, **kwargs):
        pass

    @abstractmethod
    def calculate_equilibrium_fluid_comp(self, **kwargs):
        pass

    @abstractmethod
    def calculate_saturation_pressure(self, **kwargs):
        pass

    # @abstractmethod
    # def preprocess_sample(self,**kwargs):
    #     pass

    def check_calibration_range(self, parameters, report_nonexistance=True):
        """ Checks whether the given parameters are within the ranges defined
        by the CalibrationRange objects for the model and its fugacity and
        activity models. An empty string will be returned if all parameters are
        within the calibration range. If a parameter is not within the
        calibration range, a description of the problem will be returned in the
        string.

        Parameters
        ----------
        parameters     dict
            Dictionary keys are the names of the parameters to be checked,
            e.g., pressure temperature, SiO2, etc. Values are the values of
            each parameter. A complete set need not be given.

        Returns
        -------
        str
            String description of any parameters falling outside of the
            calibration range.
        """
        s = ''
        for cr in self.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance)
        for cr in self.fugacity_model.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance)
        for cr in self.activity_model.calibration_ranges:
            if cr.check(parameters) is False:
                s += cr.string(parameters, report_nonexistance)
        return s

    def get_calibration_range(self):
        """ Returns a string describing the calibration ranges defined by the
        CalibrationRange objects for each model, and its associated fugacity
        and activity models.

        Returns
        -------
        str
            String description of the calibration range objects."""
        s = ''
        for cr in self.calibration_ranges:
            s += cr.string(None)
        for cr in self.fugacity_model.calibration_ranges:
            s += cr.string(None)
        for cr in self.activity_model.calibration_ranges:
            s += cr.string(None)
        return s


# ------------ MIXED FLUID MODELS ------------------------------- #
class MixedFluid(Model):
    """
    Implements the generic framework for mixed fluid solubility. Any set of
    pure fluid solubility models may be specified.
    """
    def __init__(self, models):
        """
        Initializes the mixed fluid model.

        Parameters
        ----------
        models     dictionary
            Dictionary with names of volatile species as keys, and the model
            objects as values.
        """
        self.models = tuple(model for model in models.values())
        self.set_volatile_species(list(models.keys()))

    def get_calibration_values(self, variable_names):
        """ Placeholder method to prevent an error when this generic method is
        called for a MixedFluid model.

        Returns
        -------
        np.nan
        """
        return np.nan

    def calculate_dissolved_volatiles(self, pressure, X_fluid,
                                      returndict=False, **kwargs):
        """
        Calculates the dissolved volatile concentrations in wt%, using each
        model's calculate_dissolved_volatiles method. At present the volatile
        concentrations are not propagated through.

        Parameters
        ----------
        pressure     float
            The total pressure in bars.
        X_fluid     float, numpy.ndarry, dict, pandas Series
            The mole fraction of each species in the fluid. If the mixed fluid
            model contains only two species (e.g. CO2 and H2O), the value of
            the first species in self.volatile_species may be passed on its own
            as a float.
        returndict         bool
            If True, the results will be returned in a dict, otherwise they
            will be returned as a tuple.

        Returns
        -------
        tuple
            Dissolved volatile concentrations of each species in the model, in
            the order set by self.volatile_species.
        """
        if ((type(X_fluid) == float or type(X_fluid) == int) and
           len(self.volatile_species) == 2):
            X_fluid = (X_fluid, 1-X_fluid)
        elif len(X_fluid) != len(self.volatile_species):
            raise core.InputError("X_fluid must have the same length as the "
                                  "number of volatile species in the "
                                  "MixedFluids Model class, or it may have "
                                  "length 1 if two species are present in "
                                  "the MixedFluids Model class.")

        if np.sum(X_fluid) != 1.0:
            raise core.InputError("X_fluid must sum to 1.0")
        if any(val < 0 for val in X_fluid) or any(val > 1 for val in X_fluid):
            raise core.InputError("Each mole fraction in X_fluid must have a "
                                  "value between 0 and 1.")

        if type(X_fluid) == dict or type(X_fluid) == pd.core.series.Series:
            X_fluid = tuple(X_fluid[species] for species in
                            self.volatile_species)

        # If the models don't depend on the concentration of volatiles,
        # themselves:
        if all(model.solubility_dependence is False for model in self.models):
            result = tuple(model.calculate_dissolved_volatiles(
                        pressure=pressure, X_fluid=Xi, **kwargs) for model, Xi
                        in zip(self.models, X_fluid))
        # If one of the models depends on the other volatile concentration
        elif (len(self.models) == 2 and
              self.models[0].solubility_dependence is False and
              'sample' in kwargs):
            result0 = self.models[0].calculate_dissolved_volatiles(
                              pressure=pressure, X_fluid=X_fluid[0], **kwargs)
            samplecopy = kwargs['sample'].change_composition(
                             {self.volatile_species[0]: result0},
                             inplace=False)
            kwargs['sample'] = samplecopy
            result1 = self.models[1].calculate_dissolved_volatiles(
                            pressure=pressure, X_fluid=X_fluid[1], **kwargs)
            result = (result0, result1)
        elif(len(self.models) == 2 and
             self.models[1].solubility_dependence is False and
             'sample' in kwargs):
            result1 = self.models[1].calculate_dissolved_volatiles(
                               pressure=pressure, X_fluid=X_fluid[1], **kwargs)
            samplecopy = kwargs['sample'].change_composition(
                            {self.volatile_species[1]: result1}, inplace=False)
            kwargs['sample'] = samplecopy
            result0 = self.models[0].calculate_dissolved_volatiles(
                               pressure=pressure, X_fluid=X_fluid[0], **kwargs)
            result = (result0, result1)
        else:
            raise core.InputError("The solubility dependence of the models "
                                  "is not currently supported by the "
                                  "MixedFluid model.")

        if returndict:
            resultsdict = {}
            for i, v in zip(range(len(self.volatile_species)),
                            self.volatile_species):
                resultsdict.update({v+'_liq': result[i]})
            return resultsdict
        else:
            return result

    def calculate_equilibrium_fluid_comp(self, pressure, sample,
                                         return_dict=True, **kwargs):
        """ Calculates the composition of the fluid in equilibrium with the
        dissolved volatile concentrations passed. If a fluid phase is
        undersaturated at the chosen pressure (0,0) will be returned. Note,
        this currently assumes the given H2O and CO2 concentrations are the
        system total, not the total dissolved. If one of the volatile species
        has a zero or negative concentration, the pure fluid model for the
        other volatile species will be used.

        Parameters
        ----------
        pressure     float
            The total pressure in bars.
        sample    Sample class
            Magma major element composition.
        return_dict     bool
            Set the return type, if true a dict will be returned, if False two
            floats will be returned. Default is True.

        Returns
        -------
        dict or floats
            Mole fractions of the volatile species in the fluid, in the order
            given by self.volatile_species if floats.
        """
        if len(self.volatile_species) != 2:
            raise core.InputError("Currently equilibrium fluid compositions "
                                  "can only be calculated when two volatile "
                                  "species are present.")

        dissolved_at_0bar = [self.models[0].calculate_dissolved_volatiles(
                                    sample=sample, pressure=0.0, **kwargs),
                             self.models[1].calculate_dissolved_volatiles(
                                    sample=sample, pressure=0.0, **kwargs)]

        if (sample.get_composition(self.volatile_species[0]) <= 0.0 or
           (sample.get_composition(self.volatile_species[0]) <=
           dissolved_at_0bar[0])):
            Xv0 = 0.0
            Xv1 = self.models[1].calculate_equilibrium_fluid_comp(
                                    pressure=pressure, sample=sample, **kwargs)
        elif (sample.get_composition(self.volatile_species[1]) <= 0.0 or
              (sample.get_composition(self.volatile_species[1]) <=
              dissolved_at_0bar[1])):
            Xv1 = 0.0
            Xv0 = self.models[0].calculate_equilibrium_fluid_comp(
                                    pressure=pressure, sample=sample, **kwargs)
        else:
            satP = self.calculate_saturation_pressure(sample, **kwargs)

            if satP < pressure:
                if return_dict:
                    return {self.volatile_species[0]: 0,
                            self.volatile_species[1]: 0}
                else:
                    return (0, 0)

            molfracs = sample.get_composition(units='mol_oxides')
            (Xt0, Xt1) = (molfracs[self.volatile_species[0]],
                          molfracs[self.volatile_species[1]])

            try:
                Xv0 = root_scalar(self.root_for_fluid_comp,
                                  bracket=[1e-15, 1-1e-15],
                                  args=(pressure, Xt0, Xt1, sample,
                                        kwargs)).root
                Xv1 = 1 - Xv0
            except Exception:
                try:
                    Xv0 = root_scalar(self.root_for_fluid_comp, x0=0.5, x1=0.1,
                                      args=(pressure, Xt0, Xt1, sample,
                                            kwargs)).root
                    Xv1 = 1 - Xv0
                except Exception:
                    raise core.SaturationError("Equilibrium fluid not found. "
                                               "Likely an issue with the "
                                               "numerical solver.")

        if return_dict:
            return {self.volatile_species[0]: Xv0,
                    self.volatile_species[1]: Xv1}
        else:
            return Xv0, Xv1

    def calculate_saturation_pressure(self, sample, **kwargs):
        """
        Calculates the pressure at which a fluid will be saturated, given the
        dissolved volatile concentrations. If one of the volatile species has a
        zero or negative concentration the pure fluid model for the other
        species will be used. If one of the volatile species has a
        concentration lower than the concentration dissolved at 0 bar, the pure
        fluid model for the other species will be used.

        Parameters
        ----------
        sample     Sample class
            Magma major element composition (including volatiles).

        Returns
        -------
        float
            The saturation pressure in bars.
        """
        dissolved_at_0bar = [self.models[0].calculate_dissolved_volatiles(
                                        sample=sample, pressure=0.0, **kwargs),
                             self.models[1].calculate_dissolved_volatiles(
                                        sample=sample, pressure=0.0, **kwargs)]

        if (sample.get_composition(self.volatile_species[0]) <= 0.0 or
            (sample.get_composition(self.volatile_species[0]) <=
             dissolved_at_0bar[0])):
            satP = self.models[1].calculate_saturation_pressure(sample=sample,
                                                                **kwargs)
        elif (sample.get_composition(self.volatile_species[1]) <= 0.0 or
              (sample.get_composition(self.volatile_species[1]) <=
               dissolved_at_0bar[1])):
            satP = self.models[0].calculate_saturation_pressure(sample=sample,
                                                                **kwargs)
        else:
            volatile_concs = np.array(tuple(sample.get_composition(species) for
                                            species in self.volatile_species))

            x0 = 0
            for model in self.models:
                xx0 = model.calculate_saturation_pressure(sample=sample,
                                                          **kwargs)
                if np.isfinite(xx0):
                    x0 += xx0

            try:
                satP = root(self.root_saturation_pressure, x0=[x0, 0.5],
                            args=(volatile_concs, sample, kwargs)).x[0]
            except Exception:
                w.warn("Saturation pressure not found.", RuntimeWarning,
                       stacklevel=2)
                satP = np.nan

        return satP

    def calculate_isobars_and_isopleths(self, pressure_list,
                                        isopleth_list=[0, 1], points=51,
                                        return_dfs=True, extend_to_zero=True,
                                        **kwargs):
        """
        Calculates isobars and isopleths. Isobars can be calculated for any
        number of pressures. Variables required by each of the pure fluid
        models must be passed, e.g. sample, temperature, etc.

        Parameters
        ----------
        pressure_list     list or float
            List of all pressure values at which to calculate isobars, in bars.
        isopleth_list     list
            Default value is None, in which case only isobars will be
            calculated. List of all fluid compositions in mole fraction (of the
            first species in self.volatile_species) at which to calcualte
            isopleths. Values can range from 0 to 1.
        points     int
            The number of points in each isobar and isopleth. Default value is
            101.
        return_dfs     bool
            If True, the results will be returned as two pandas DataFrames, as
            produced by the MagmaSat method. If False the results will be
            returned as lists of numpy arrays.

        Returns
        -------
        pandas DataFrame object(s) or list(s)
            If isopleth_list is not None, two objects will be returned, one
            with the isobars and the second withthe isopleths. If return_dfs is
            True, two pandas DataFrames will be returned with column names
            'Pressure' or 'XH2O_fl', 'H2O_liq', and 'CO2_liq'. If return_dfs is
            False, two lists of numpy arrays will be returned. Each array is an
            individual isobar or isopleth, in the order passed via
            pressure_list or isopleth_list. The arrays are the concentrations
            of H2O and CO2 in the liquid, in the order of the species in
            self.volatile_species.

        """
        if (len(self.volatile_species) != 2 or
                'H2O' not in self.volatile_species or
                'CO2' not in self.volatile_species):
            raise core.InputError("calculate_isobars_and_isopleths may only "
                                  "be used with a H2O-CO2 fluid.")

        H2O_id = self.volatile_species.index('H2O')
        CO2_id = self.volatile_species.index('CO2')

        if isinstance(pressure_list, list):
            pass
        elif (isinstance(pressure_list, int) or
              isinstance(pressure_list, float)):
            pressure_list = [pressure_list]
        else:
            raise core.InputError("pressure_list must be a single float "
                                  "(1000.0), int (1000), or list of those "
                                  "[1000, 2000.0, 3000].")

        has_isopleths = True
        if isopleth_list is None:
            has_isopleths = False

        isobars_df = pd.DataFrame(columns=['Pressure', 'H2O_liq', 'CO2_liq'])
        isobars = []
        for pressure in pressure_list:
            dissolved = np.zeros([2, points])
            Xv0 = np.linspace(0.0, 1.0, points)
            for i in range(points):
                dissolved[:, i] = self.calculate_dissolved_volatiles(
                       pressure=pressure, X_fluid=(Xv0[i], 1-Xv0[i]), **kwargs)
                isobars_df = isobars_df.append({
                                               'Pressure': pressure,
                                               'H2O_liq': dissolved[H2O_id, i],
                                               'CO2_liq': dissolved[CO2_id, i]
                                               }, ignore_index=True)
            isobars.append(dissolved)

        if has_isopleths:
            isopleths_df = pd.DataFrame(columns=['XH2O_fl', 'H2O_liq',
                                                 'CO2_liq'])
            isopleths = []
            for isopleth in isopleth_list:
                dissolved = np.zeros([2, points])
                pmin = np.nanmin(pressure_list)
                pmax = np.nanmax(pressure_list)
                if pmin == pmax:
                    pmin = 0.0
                pressure = np.linspace(pmin, pmax, points)
                for i in range(points):
                    dissolved[:, i] = self.calculate_dissolved_volatiles(
                          pressure=pressure[i], X_fluid=(isopleth, 1-isopleth),
                          **kwargs)
                    isopleths_df = isopleths_df.append({
                       'XH2O_fl': [isopleth, 1-isopleth][H2O_id],
                       'H2O_liq': dissolved[H2O_id, i],
                       'CO2_liq': dissolved[CO2_id, i]}, ignore_index=True)
                isopleths.append(dissolved)

        if return_dfs:
            if has_isopleths:
                return (isobars_df, isopleths_df)
            else:
                return isobars_df
        else:
            if has_isopleths:
                return (isobars, isopleths)
            else:
                return isobars

    def calculate_degassing_path(self, sample, pressure='saturation',
                                 fractionate_vapor=0.0, final_pressure=100.0,
                                 steps=101, return_dfs=True,
                                 round_to_zero=True, **kwargs):
        """
        Calculates the dissolved volatiles in a progressively degassing sample.

        Parameters
        ----------
        sample     Sample class
            Magma major element composition (including volatiles).
        pressure     string, float, int, list, or numpy array
            Defaults to 'saturation', the calculation will begin at the
            saturation pressure. If a number is passed as either a float or
            int, this will be the starting pressure. If a list of numpy array
            is passed, the pressure values in the list or array will define the
            degassing path, i.e. final_pressure and steps variables will be
            ignored. Units are bars.
        fractionate_vapor     float
            What proportion of vapor should be removed at each step. If 0.0
            (default), the degassing path will correspond to closed-system
            degassing. If 1.0, the degassing path will correspond to
            open-system degassing.
        final_pressure         float
            The final pressure on the degassing path, in bars. Ignored if a
            list or numpy array is passed as the pressure variable. Default is
            1 bar.
        steps     int
            The number of steps in the degassing path. Ignored if a list or
            numpy array are passed as the pressure variable.
        return_dfs     bool
            If True, the results will be returned in a pandas DataFrame, if
            False, two numpy arrays will be returned.
        round_to_zero   bool
            If True, the first entry of FluidProportion_wt will be rounded to
            zero, rather than being a value within numerical error of zero.
            Default is True.

        Returns
        -------
        pandas DataFrame or numpy arrays
            If return_dfs is True (default), a DataFrame with columns
            'Pressure_bars', 'H2O_liq', 'CO2_liq', 'H2O_fl', 'CO2_fl', and
            'FluidProportion_wt', is returned. Dissolved volatiles are in wt%,
            the proportions of volatiles in the fluid are in mole fraction.
            Otherwise a numpy array containing the dissolved volatile
            concentrations, and a numpy array containing the mole fractions of
            volatiles in the fluid is returned. The columns are in the order of
            the volatiles in self.volatile_species.
        """

        # Create a copy of the sample so that initial volatile concentrations
        # are not overwritten.
        sample = deepcopy(sample)

        # Its imperative that normalization doesn't change the volatile
        # concentrations throughout the calculation.
        if sample.default_normalization not in ['fixedvolatiles', 'none']:
            sample.set_default_normalization('fixedvolatiles')
            w.warn('Sample normalization changed to fixedvolatiles.')

        wtptoxides = sample.get_composition(units='wtpt_oxides')
        wtm0s, wtm1s = (wtptoxides[self.volatile_species[0]],
                        wtptoxides[self.volatile_species[1]])

        if pressure == 'saturation':
            p0 = self.calculate_saturation_pressure(sample, **kwargs)
            pressures = np.linspace(p0, final_pressure, steps)
        elif type(pressure) == float or type(pressure) == int:
            pressures = np.linspace(pressure, final_pressure, steps)
        elif type(pressure) == list or type(pressure) == np.ndarray:
            pressures = pressure

        Xv = np.zeros([2, len(pressures)])
        wtm = np.zeros([2, len(pressures)])

        for i in range(len(pressures)):
            try:
                wtptoxides = sample.get_composition(units='wtpt_oxides')
                X_fluid = self.calculate_equilibrium_fluid_comp(
                    pressure=pressures[i], sample=sample, return_dict=False,
                    **kwargs)
                Xv[:, i] = X_fluid
                if X_fluid == (0, 0):
                    wtm[:, i] = (wtptoxides[self.volatile_species[0]],
                                 wtptoxides[self.volatile_species[1]])
                else:
                    if X_fluid[0] == 0:
                        wtm[0, i] = wtptoxides[self.volatile_species[0]]
                        wtm[1, i] = self.calculate_dissolved_volatiles(
                            pressure=pressures[i], sample=sample,
                            X_fluid=X_fluid, **kwargs)[1]
                    elif X_fluid[1] == 0:
                        wtm[1, i] = wtptoxides[self.volatile_species[1]]
                        wtm[0, i] = self.calculate_dissolved_volatiles(
                            pressure=pressures[i], sample=sample,
                            X_fluid=X_fluid, **kwargs)[0]
                    else:
                        wtm[:, i] = self.calculate_dissolved_volatiles(
                            pressure=pressures[i], sample=sample,
                            X_fluid=X_fluid, **kwargs)

                    sample.change_composition({
                        self.volatile_species[0]: (wtm[0, i] +
                                                   (1-fractionate_vapor) *
                                                   (wtm0s-wtm[0, i])),
                        self.volatile_species[1]: (wtm[1, i] +
                                                   (1-fractionate_vapor) *
                                                   (wtm1s-wtm[1, i]))})

            except Exception:
                Xv[:, i] = [np.nan]*np.shape(Xv)[0]
                wtm[:, i] = wtm[:, i-1]

        if return_dfs:
            exsolved_degassing_df = pd.DataFrame()
            exsolved_degassing_df['Pressure_bars'] = pressures
            exsolved_degassing_df['H2O_liq'] = (
                                    wtm[self.volatile_species.index('H2O'), :])
            exsolved_degassing_df['CO2_liq'] = (
                                    wtm[self.volatile_species.index('CO2'), :])
            exsolved_degassing_df['H2O_fl'] = (
                                     Xv[self.volatile_species.index('H2O'), :])
            exsolved_degassing_df['CO2_fl'] = (
                                     Xv[self.volatile_species.index('CO2'), :])
            exsolved_degassing_df['FluidProportion_wt'] = (
                (wtm0s+wtm1s) - exsolved_degassing_df['H2O_liq'] -
                exsolved_degassing_df['CO2_liq'])

            if (round_to_zero is True and np.round(
                  exsolved_degassing_df.loc[0, 'FluidProportion_wt'], 2) == 0):
                exsolved_degassing_df.loc[0, 'FluidProportion_wt'] = 0.0

            return exsolved_degassing_df

        else:
            return (wtm, Xv)

    def root_saturation_pressure(self, x, volatile_concs, sample, kwargs):
        """ Function called by scipy.root when finding the saturation pressure
        using calculate_saturation_pressure.

        Parameters
        ----------
        x     numpy array
            The guessed value for the root. x[0] is the pressure (in bars) and
            x[1] is the mole fraction of the first volatile in
            self.volatile_species.
        volatile_concs     numpy array
            The dissolved volatile concentrations, in the same order as
            self.volatile_species.
        sample:     Sample class
            Magma major element composition (including volatiles).
        kwargs     dictionary
            Dictionary of keyword arguments, which may be required by the
            pure-fluid models.

        Returns
        -------
        numpy array
            The difference in the dissolved volatile concentrations, and those
            predicted with the pressure and fluid composition specified by x.
        """
        if x[1] < 0:
            x[1] = 0
        elif x[1] > 1:
            x[1] = 1
        if x[0] <= 0:
            x[0] = 1e-15
        misfit = (np.array(self.calculate_dissolved_volatiles(
            pressure=x[0], X_fluid=(x[1], 1-x[1]), sample=sample, **kwargs))
            - volatile_concs)
        return misfit

    def root_for_fluid_comp(self, Xv0, pressure, Xt0, Xt1, sample, kwargs):
        """ Function called by scipy.root_scalar when calculating the
        composition of equilibrium fluid in the
        calculate_equilibrium_fluid_comp method.

        Parameters
        ----------
        Xv0     float
            The guessed mole fraction of the first volatile species in
            self.volatile_species.
        pressure     float
            The total pressure in bars.
        Xt0     float
            The total mole fraction of the first volatile species in
            self.volatile_species.
        Xt1        float
            The total mole fraction of the second volatile species in
            self.volatile_species.
        sample     Sample class
            Magma major element composition.
        kwargs     dictionary
            A dictionary of keyword arguments that may be required by the pure
            fluid models.

        Returns
        -------
        float
            The differene in the LHS and RHS of the mass balance equation. Eq X
            in manuscript.
        """

        wtt0 = sample.get_composition(self.volatile_species[0])
        wtt1 = sample.get_composition(self.volatile_species[1])

        wtm0, wtm1 = self.calculate_dissolved_volatiles(
            pressure=pressure, X_fluid=(Xv0, 1-Xv0), sample=sample, **kwargs)

        Xm0 = Xt0 / wtt0 * wtm0
        Xm1 = Xt1 / wtt1 * wtm1

        if self.volatile_species[0] == 'CO2' and Xv0 != Xm0:
            f = (Xt0 - Xm0) / (Xv0 - Xm0)
            return (1 - f) * Xm1 + f * (1 - Xv0) - Xt1
        else:
            f = (Xt1 - Xm1) / ((1 - Xv0) - Xm1)
            return (1 - f) * Xm0 + f * Xv0 - Xt0

    def check_calibration_range(self, parameters, report_nonexistance=True):
        """ Checks whether the given parameters are within the ranges defined
        by the CalibrationRange objects for each model and its fugacity and
        activity models. An empty string will be returned if all parameters are
        within the calibration range. If a parameter is not within the
        calibration range, a description of the problem will be returned in the
        string.

        Parameters
        ----------
        parameters     dict
            Dictionary keys are the names of the parameters to be checked,
            e.g., pressure temperature, SiO2, etc. Values are the values of
            each parameter. A complete set need not be given.

        Returns
        -------
        str
            String description of any parameters falling outside of the
            calibration range.
        """
        s = ''
        for model in self.models:
            for cr in model.calibration_ranges:
                if cr.check(parameters) is False:
                    s += cr.string(parameters, report_nonexistance)
            for cr in model.fugacity_model.calibration_ranges:
                if cr.check(parameters) is False:
                    s += cr.string(parameters, report_nonexistance)
            for cr in model.activity_model.calibration_ranges:
                if cr.check(parameters) is False:
                    s += cr.string(parameters, report_nonexistance)
        return s

    def get_calibration_range(self):
        """ Returns a string describing the calibration ranges defined by the
        CalibrationRange objects for each model, and its associated fugacity
        and activity models.

        Returns
        -------
        str
            String description of the calibration range objects."""
        s = ''
        for model in self.models:
            for cr in model.calibration_ranges:
                s += cr.string(None)
            for cr in model.fugacity_model.calibration_ranges:
                s += cr.string(None)
            for cr in model.activity_model.calibration_ranges:
                s += cr.string(None)
        return s



import unittest
import VESIcal as v
import numpy as np
import pandas as pd

class TestCreateSample(unittest.TestCase):
    def setUp(self):
        self.majors = pd.Series({'SiO2':    47.95,
                                 'TiO2':    1.67,
                                 'Al2O3':   17.32,
                                 'FeO':     10.24,
                                 'Fe2O3':   0.1,
                                 'MgO':     5.76,
                                 'CaO':     10.93,
                                 'Na2O':    3.45,
                                 'K2O':     1.99,
                                 'P2O5':    0.51,
                                 'MnO':     0.1,
                    })

        self.majors_normed = pd.Series({'SiO2':    47.94,
                                        'TiO2':    1.67,
                                        'Al2O3':   17.32,
                                        'FeO':     10.24,
                                        'Fe2O3':   0.1,
                                        'MgO':     5.76,
                                        'CaO':     10.93,
                                        'Na2O':    3.45,
                                        'K2O':     1.99,
                                        'P2O5':    0.51,
                                        'MnO':     0.1,
                    })

        # Mol fractions calculated externally
        self.majors_moloxides = pd.Series({ 'SiO2':  51.43,
                                            'TiO2':  1.35,
                                            'Al2O3': 10.95,
                                            'FeO':   9.19,
                                            'Fe2O3': 0.04,
                                            'MnO':   0.09,
                                            'MgO':   9.21,
                                            'CaO':   12.56,
                                            'Na2O':  3.59,
                                            'K2O':   1.36,
                                            'P2O5': 0.23})
        self.majors_moloxides = self.majors_moloxides/100

        self.majors_molcations = pd.Series({'Si':   0.4427,
                                            'Ti':   0.0116,
                                            'Al':   0.1885,
                                            'Fe':   0.0791,
                                            'Fe3':  0.0007,
                                            'Mn':   0.0008,
                                            'Mg':   0.0793,
                                            'Ca':   0.1081,
                                            'Na':   0.0618,
                                            'K':    0.0234,
                                            'Cr':   0.0040})

        self.majorsv = pd.Series({'SiO2':   47.95,
                                  'TiO2':   1.67,
                                  'Al2O3':  17.32,
                                  'FeO':    10.24,
                                  'Fe2O3':  0.1,
                                  'MgO':    5.76,
                                  'CaO':    10.93,
                                  'Na2O':   3.45,
                                  'K2O':    1.99,
                                  'P2O5':   0.51,
                                  'MnO':    0.1,
                                  'CO2':    0.08,
                                  'H2O':    4.0
                        })

        self.majorsv_normed = pd.Series({'SiO2':   46.06,
                                         'TiO2':   1.60,
                                         'Al2O3':  16.64,
                                         'FeO':    9.84,
                                         'Fe2O3':  0.1,
                                         'MgO':    5.53,
                                         'CaO':    10.50,
                                         'Na2O':   3.31,
                                         'K2O':    1.91,
                                         'P2O5':   0.49,
                                         'MnO':    0.1,
                                         'CO2':    0.08,
                                         'H2O':    3.84
                        })

        self.majorsv_moloxides = pd.Series({'SiO2':  44.95,
                                            'TiO2':  1.178,
                                            'Al2O3': 9.568,
                                            'FeO':   8.028,
                                            'Fe2O3': 0.035,
                                            'MnO':   0.079,
                                            'MgO':   8.050,
                                            'CaO':   10.978,
                                            'Na2O':  3.135,
                                            'K2O':   1.190,
                                            'P2O5':  0.202,
                                            'H2O':   12.503,
                                            'CO2':   0.102})
        self.majorsv_moloxides = self.majorsv_moloxides/100

        self.majorsv_molcations = pd.Series({'Si':   0.35496,
                                             'Ti':   0.00930,
                                             'Al':   0.15111,
                                             'Fe':   0.06340,
                                             'Fe3':  0.00056,
                                             'Mn':   0.00063,
                                             'Mg':   0.06357,
                                             'Ca':   0.08669,
                                             'Na':   0.04952,
                                             'K':    0.01879,
                                             'P':    0.00320,
                                             'H':    0.19747,
                                             'C':    0.00081})

        self.majors_all_species = pd.Series({'SiO2':    47.95,
                                             'TiO2':    1.67,
                                             'Al2O3':   17.32,
                                             'FeO':     10.24,
                                             'Fe2O3':   0.1,
                                             'MgO':     5.76,
                                             'CaO':     10.93,
                                             'Na2O':    3.45,
                                             'K2O':     1.99,
                                             'P2O5':    0.51,
                                             'MnO':     0.1,
                                             'Cr2O3':   0.05,
                                             'NiO':     0.04,
                                             'CoO':     0.02,
                                             'H2O':     2.00,
                                             'CO2':     0.12,
                                             'F2O':     0.05
                                                })

        self.sample = v.Sample(self.majorsv)
        self.sample_all = v.Sample(self.majors_all_species)

    def test_createSample(self):
        for ox in self.majors.index:
            self.assertEqual(self.sample._composition[ox],self.majors[ox])
        for ox in self.majors_all_species.index:
            self.assertEqual(self.sample_all._composition[ox],self.majors_all_species[ox])

    def test_setdefault_noargs(self):
        self.assertEqual(self.sample.default_normalization,'none')
        self.assertEqual(self.sample.default_units,'wtpt_oxides')

    def test_setdefaults_none_wtpt(self):
        sample = v.Sample(self.majorsv, default_normalization='none',default_units='wtpt_oxides')
        self.assertEqual(sample.default_normalization,'none')
        self.assertEqual(sample.default_units,'wtpt_oxides')

    def test_setdefaults_standard_molox(self):
        sample = v.Sample(self.majorsv, default_normalization='standard',default_units='mol_oxides')
        self.assertEqual(sample.default_normalization,'standard')
        self.assertEqual(sample.default_units,'mol_oxides')

    def test_setdefaults_fixvol_molcat(self):
        sample = v.Sample(self.majorsv, default_normalization='fixedvolatiles',default_units='mol_cations')
        self.assertEqual(sample.default_normalization,'fixedvolatiles')
        self.assertEqual(sample.default_units,'mol_cations')

    def test_setdefaults_addvol_molSinO(self):
        sample = v.Sample(self.majorsv, default_normalization='additionalvolatiles',default_units='mol_singleO')
        self.assertEqual(sample.default_normalization,'additionalvolatiles')
        self.assertEqual(sample.default_units,'mol_singleO')

    def test_setdefaults_garbageNorm(self):
        with self.assertRaises(v.core.InputError):
            v.Sample(composition=self.majorsv,default_normalization='garbage')

    def test_setdefaults_garbageType(self):
        with self.assertRaises(v.core.InputError):
            v.Sample(composition=self.majorsv,default_units='garbage')

    def test_type_garbage(self):
        with self.assertRaises(v.core.InputError):
            v.Sample(composition=self.majorsv,units='garbage')

    def test_type_wtptoxides(self):
        sample = v.Sample(self.majors,units='wtpt_oxides')
        for ox in self.majors.index:
            self.assertEqual(self.sample._composition[ox],self.majors[ox])

    def test_type_moloxides(self):
        sample = v.Sample(self.majorsv_moloxides,units='mol_oxides')
        for ox in self.majorsv.index:
            self.assertEqual(np.round(sample._composition[ox],2),np.round(self.majorsv_normed[ox],2))

    def test_type_molcations(self):
        sample = v.Sample(self.majorsv_molcations,units='mol_cations')
        for ox in self.majorsv.index:
            self.assertEqual(np.round(sample._composition[ox],2),np.round(self.majorsv_normed[ox],2))


class TestGetComposition(unittest.TestCase):
    def setUp(self):
        self.majors = pd.Series({'SiO2':    47.95,
                                 'TiO2':    1.67,
                                 'Al2O3':   17.32,
                                 'FeO':     10.24,
                                 'Fe2O3':   0.1,
                                 'MgO':     5.76,
                                 'CaO':     10.93,
                                 'Na2O':    3.45,
                                 'K2O':     1.99,
                                 'P2O5':    0.51,
                                 'MnO':     0.1,
                    })

        self.majors_normed = pd.Series({'SiO2':    47.94,
                                        'TiO2':    1.67,
                                        'Al2O3':   17.32,
                                        'FeO':     10.24,
                                        'Fe2O3':   0.1,
                                        'MgO':     5.76,
                                        'CaO':     10.93,
                                        'Na2O':    3.45,
                                        'K2O':     1.99,
                                        'P2O5':    0.51,
                                        'MnO':     0.1,
                    })

        # Mol fractions calculated externally
        self.majors_moloxides = pd.Series({ 'SiO2':  51.434,
                                            'TiO2':  1.348,
                                            'Al2O3': 10.948,
                                            'FeO':   9.186,
                                            'Fe2O3': 0.004,
                                            'MnO':   0.091,
                                            'MgO':   9.211,
                                            'CaO':   12.562,
                                            'Na2O':  3.588,
                                            'K2O':   1.362,
                                            'P2O5': 0.232})
        self.majors_moloxides = self.majors_moloxides/100

        self.majors_molcations = pd.Series({'Si':   0.44275,
                                            'Ti':   0.01160,
                                            'Al':   0.18848,
                                            'Fe':   0.07908,
                                            'Fe3':  0.00069,
                                            'Mn':   0.00078,
                                            'Mg':   0.07929,
                                            'Ca':   0.10813,
                                            'Na':   0.06176,
                                            'K':    0.02344,
                                            'P':    0.00399})

        self.majors_molSingleO = pd.Series({ 'Si':   0.29276,
                                             'Ti':   0.00767,
                                             'Al':   0.12463,
                                             'Fe':   0.05229,
                                             'Fe3':  0.00046,
                                             'Mn':   0.00052,
                                             'Mg':   0.05243,
                                             'Ca':   0.07150,
                                             'Na':   0.04084,
                                             'K':    0.01550,
                                             'P':    0.00264,
                                            })

        self.majors_fw = 36.6928

        self.majorsv = pd.Series({'SiO2':   47.95,
                                  'TiO2':   1.67,
                                  'Al2O3':  17.32,
                                  'FeO':    10.24,
                                  'Fe2O3':  0.1,
                                  'MgO':    5.76,
                                  'CaO':    10.93,
                                  'Na2O':   3.45,
                                  'K2O':    1.99,
                                  'P2O5':   0.51,
                                  'MnO':    0.1,
                                  'CO2':    0.08,
                                  'H2O':    4.0
                        })

        self.majorsv_normed = pd.Series({'SiO2':   46.06,
                                         'TiO2':   1.60,
                                         'Al2O3':  16.64,
                                         'FeO':    9.84,
                                         'Fe2O3':  0.1,
                                         'MgO':    5.53,
                                         'CaO':    10.50,
                                         'Na2O':   3.31,
                                         'K2O':    1.91,
                                         'P2O5':   0.49,
                                         'MnO':    0.1,
                                         'CO2':    0.08,
                                         'H2O':    3.84
                        })

        self.majorsv_moloxides = pd.Series({'SiO2':  44.95,
                                            'TiO2':  1.178,
                                            'Al2O3': 9.568,
                                            'FeO':   8.028,
                                            'Fe2O3': 0.035,
                                            'MnO':   0.079,
                                            'MgO':   8.050,
                                            'CaO':   10.978,
                                            'Na2O':  3.135,
                                            'K2O':   1.190,
                                            'P2O5':  0.202,
                                            'H2O':   12.503,
                                            'CO2':   0.102})
        self.majorsv_moloxides = self.majorsv_moloxides/100

        self.majorsv_molcations = pd.Series({'Si':   0.35496,
                                             'Ti':   0.00930,
                                             'Al':   0.15111,
                                             'Fe':   0.06340,
                                             'Fe3':  0.00056,
                                             'Mn':   0.00063,
                                             'Mg':   0.06357,
                                             'Ca':   0.08669,
                                             'Na':   0.04952,
                                             'K':    0.01879,
                                             'P':    0.00320,
                                             'H':    0.19747,
                                             'C':    0.00081})

        self.majorsv_molSingleO = pd.Series({'Si':   0.27038,
                                             'Ti':   0.00708,
                                             'Al':   0.11510,
                                             'Fe':   0.04829,
                                             'Fe3':  0.00042,
                                             'Mn':   0.00048,
                                             'Mg':   0.04842,
                                             'Ca':   0.06604,
                                             'Na':   0.03772,
                                             'K':    0.01432,
                                             'P':    0.00243,
                                             'H':    0.15042,
                                             'C':    0.00062})

        self.majorsv_fw = 35.2704

        self.sample = v.Sample(self.majorsv)

    def test_default(self):
        composition = self.sample.get_composition()
        for ox in composition.index:
            self.assertEqual(composition[ox],self.majorsv[ox])

    def test_wtptoxides_none(self):
        composition = self.sample.get_composition(normalization='none')
        for ox in composition.index:
            self.assertEqual(composition[ox],self.majorsv[ox])

    def test_wtptoxides_none_exclV(self):
        composition = self.sample.get_composition(normalization='none',exclude_volatiles=True)
        for ox in composition.index:
            self.assertEqual(composition[ox],self.majors[ox])

    def test_wtptoxides_std(self):
        composition = self.sample.get_composition(normalization='standard')
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],2),np.round(self.majorsv_normed[ox],2))

    def test_wtptoxides_std_exclV(self):
        composition = self.sample.get_composition(normalization='standard',exclude_volatiles=True)
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],2),np.round(self.majors_normed[ox],2))

    def test_wtptoxides_fixedVolatiles(self):
        composition = self.sample.get_composition(normalization='fixedvolatiles')
        for ox in composition.index:
            if ox not in ['CO2','H2O']:
                self.assertEqual(np.round(composition[ox],2),np.round(self.majors_normed[ox]*(100-self.majorsv['CO2']-self.majorsv['H2O'])/100,2))
            else:
                self.assertEqual(np.round(composition[ox],2),np.round(self.majorsv[ox],2))

    def test_wtptoxides_fixedVolatiles_exclV(self):
        composition = self.sample.get_composition(normalization='fixedvolatiles',exclude_volatiles=True)
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],2),np.round(self.majors_normed[ox],2))

    def test_wtptoxides_additionalVolatiles(self):
        composition = self.sample.get_composition(normalization='additionalvolatiles')
        for ox in composition.index:
            if ox not in ['CO2','H2O']:
                self.assertEqual(np.round(composition[ox],2),np.round(self.majors_normed[ox],2))
            else:
                self.assertEqual(np.round(composition[ox],2),np.round(self.majorsv[ox],2))

    def test_wtptoxides_additionalVolatiles_exclV(self):
        composition = self.sample.get_composition(normalization='additionalvolatiles',exclude_volatiles=True)
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],2),np.round(self.majors_normed[ox],2))

    def test_moloxides(self):
        composition = self.sample.get_composition(units='mol_oxides')
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],3),np.round(self.majorsv_moloxides[ox],3))

    def test_moloxides_exclV(self):
        composition = self.sample.get_composition(units='mol_oxides',exclude_volatiles=True)
        for ox in composition.index:
            self.assertEqual(np.round(composition[ox],3),np.round(self.majors_moloxides[ox],3))

    def test_molcations(self):
        composition = self.sample.get_composition(units='mol_cations')
        for el in composition.index:
            self.assertEqual(np.round(composition[el],3),np.round(self.majorsv_molcations[el],3))

    def test_molcations_exclV(self):
        composition = self.sample.get_composition(units='mol_cations',exclude_volatiles=True)
        for el in composition.index:
            self.assertEqual(np.round(composition[el],3),np.round(self.majors_molcations[el],3))

    def test_molsingleO(self):
        composition = self.sample.get_composition(units='mol_singleO')
        for el in composition.index:
            self.assertEqual(np.round(composition[el],3),np.round(self.majorsv_molSingleO[el],3))

    def test_molsingleO_exclV(self):
        composition = self.sample.get_composition(units='mol_singleO',exclude_volatiles=True)
        for el in composition.index:
            self.assertEqual(np.round(composition[el],3),np.round(self.majors_molSingleO[el],3))

    def test_formulawt(self):
        fw = self.sample.get_formulaweight()
        self.assertEqual(np.round(self.majorsv_fw,2),np.round(fw,2))

    def test_formulawt_exclV(self):
        fw = self.sample.get_formulaweight(exclude_volatiles=True)
        self.assertEqual(np.round(self.majors_fw,2),np.round(fw,2))



from setuptools import setup

install_requires = [
    'matplotlib'
    ]

setup(
    name='TASplot',
    version=0.1,
    description="""
        tasplot is a Python module that adds the names of volcanic rock
        types to a plot of total alkali versus silica data""",
    license='MIT',
    author='John A. Stevenson',
    author_email='@volcan01010',
    install_requires=install_requires,
    packages=['tasplot'],
    classifiers=[
        'Development Status :: 3 - Beta',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ]
)



from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
import warnings as w
import sympy
from scipy.optimize import root_scalar


class water(model_classes.Model):
    """
    Implementation of the Liu et al. (2005) H2O solubility model for metaluminous high-silica
    rhyolitic melts.
    """

    def __init__(self):
        """
        Initialize the model.
        """
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)

        # Generate calibration range objects for each oxide
        cr_oxide_list = []
        for ox in watercomprange.keys():
            cr_oxide_list.append(
                calibration_checks.CalibrationRange(
                    ox, watercomprange[ox], calibration_checks.crf_Between, 'wt%',
                    'Liu et al. (2005) water',
                    fail_msg=calibration_checks.crmsg_BC_fail,
                    pass_msg=calibration_checks.crmsg_BC_pass,
                    description_msg=calibration_checks.crmsg_Between_description))

        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [0, 5000.0], calibration_checks.crf_Between, 'bar',
                'Liu et al. (2005) water',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [700.0, 1200], calibration_checks.crf_Between, 'oC',
                'Liu et al. (2005) water',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'sample', None, crf_WaterComp, None, None,
                fail_msg=crmsg_WaterComp_fail,
                pass_msg=crmsg_Comp_pass,
                description_msg=crmsg_Comp_description)] + cr_oxide_list)

    def calculate_dissolved_volatiles(self, sample, pressure, temperature, X_fluid=1.0, **kwargs):
        """
        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 1.0. Mole fraction of H2O in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated dissolved H2O concentration in wt%.
        """
        pressureMPa = pressure / 10.0
        Pw = pressureMPa * X_fluid
        PCO2 = pressureMPa * (1 - X_fluid)

        temperatureK = temperature + 273.15

        H2Ot = ((354.94*Pw**(0.5) + 9.623*Pw - 1.5223*Pw**(1.5)) / temperatureK +
                0.0012439*Pw**(1.5) + PCO2*(-1.084*10**(-4)*Pw**(0.5) - 1.362*10**(-5)*Pw))

        return H2Ot

    def calculate_equilibrium_fluid_comp(self, sample, pressure, temperature, **kwargs):
        """
        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        Returns
        -------
        float
            Calculated equilibrium fluid concentration in XH2Ofluid mole fraction.
        """
        temperatureK = temperature + 273.15
        pressureMPa = pressure / 10.0

        H2Ot = sample.get_composition("H2O")

        # calculate saturation pressure and assert that input P <= SatP
        satP = self.calculate_saturation_pressure(temperature, sample)
        is_saturated = satP - pressure
        if is_saturated >= 0:
            pass
        else:
            w.warn("{:.1f} bars is above the saturation pressure ({:.1f} bars) for this sample. "
                   "Results from this calculation may be nonsensical.".format(pressure, satP))

        # Use sympy to solve solubility equation for XH2Ofluid
        XH2Ofluid = sympy.symbols('XH2Ofluid')  # XH2Ofluid is the variable to solve for

        equation = ((354.94*(XH2Ofluid*pressureMPa)**(0.5) + 9.623*(XH2Ofluid*pressureMPa)
                    - 1.5223*(XH2Ofluid*pressureMPa)**(1.5)) / temperatureK
                    + 0.0012439*(XH2Ofluid*pressureMPa)**(1.5)
                    + pressureMPa*(1-XH2Ofluid)*(-1.084*10**(-4)*(XH2Ofluid*pressureMPa)**(0.5)
                    - 1.362*10**(-5)*(XH2Ofluid*pressureMPa)) - H2Ot)

        XH2Ofluid = sympy.solve(equation, XH2Ofluid)[0]
        if XH2Ofluid > 1:
            XH2Ofluid = 1
        if XH2Ofluid < 0:
            XH2Ofluid = 0

        return XH2Ofluid

    def calculate_saturation_pressure(self, temperature, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a an H2O-bearing fluid is saturated. Calls the
        scipy.root_scalar routine, which makes repeated called to the
        calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 1.0. Mole fraction of H2O in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        temperatureK = temperature + 273.15
        if temperatureK <= 0.0:
            raise core.InputError("Temperature must be greater than 0K.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('H2O') is False:
            raise core.InputError("sample must contain H2O.")
        if sample.get_composition('H2O') < 0.0:
            raise core.InputError("Dissolved H2O concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure,
                               args=(temperature, sample, X_fluid, kwargs),
                               x0=1.0, x1=2.0).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return np.real(satP)

    def root_saturation_pressure(self, pressure, temperature, sample, X_fluid, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample:    Sample class
            Magma major element composition, including H2O.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, X_fluid=X_fluid, **kwargs) -
                sample.get_composition('H2O'))


class carbon(model_classes.Model):
    """
    Implementation of the Liu et al. (2005) H2O-CO2 solubility model for metaluminous high-silica
    rhyolitic melts.
    """

    def __init__(self):
        """
        Initialize the model.
        """
        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)

        # Generate calibration range objects for each oxide
        cr_oxide_list = []
        for ox in carboncomprange.keys():
            cr_oxide_list.append(
                calibration_checks.CalibrationRange(
                    ox, carboncomprange[ox], calibration_checks.crf_Between, 'wt%',
                    'Liu et al. (2005) carbon',
                    fail_msg=calibration_checks.crmsg_BC_fail,
                    pass_msg=calibration_checks.crmsg_BC_pass,
                    description_msg=calibration_checks.crmsg_Between_description))

        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [0, 5000.0], calibration_checks.crf_Between, 'bar',
                'Liu et al. (2005) Carbon',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [700.0, 1200], calibration_checks.crf_Between, 'oC',
                'Liu et al. (2005) Carbon',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'sample', None, crf_CarbonComp, None, None,
                fail_msg=crmsg_CarbonComp_fail,
                pass_msg=crmsg_Comp_pass,
                description_msg=crmsg_Comp_description)] + cr_oxide_list)

    def calculate_dissolved_volatiles(self, sample, pressure, temperature, X_fluid=1, **kwargs):
        """
        Parameters
        ----------
        sample:        Sample class
            Magma major element composition.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 1. Mole fraction of CO2 in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated dissolved CO2 concentration in wt%.
        """
        pressureMPa = pressure / 10.0
        Pw = pressureMPa * (1 - X_fluid)
        PCO2 = pressureMPa * X_fluid

        temperatureK = temperature + 273.15

        CO2melt_ppm = (PCO2*(5668 - 55.99*Pw)/temperatureK
                       + PCO2*(0.4133*Pw**(0.5) + 2.041*10**(-3)*Pw**(1.5)))

        CO2melt = CO2melt_ppm / 10000

        return CO2melt

    def calculate_equilibrium_fluid_comp(self, sample, pressure, temperature, **kwargs):
        """
        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        Returns
        -------
        float
            Calculated equilibrium fluid concentration in XCO2fluid mole fraction.
        """
        temperatureK = temperature + 273.15
        pressureMPa = pressure / 10.0

        if temperatureK <= 0.0:
            raise core.InputError("Temperature must be greater than 0K.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0.0:
            raise core.InputError("Dissolved CO2 concentration must be greater than 0 wt%.")

        CO2melt_wt = sample.get_composition("CO2")
        CO2melt_ppm = CO2melt_wt * 10000

        # calculate saturation pressure and assert that input P <= SatP
        satP = self.calculate_saturation_pressure(temperature, sample)
        is_saturated = satP - pressure
        if is_saturated >= 0:
            pass
        else:
            w.warn(str(pressure) + " bars is above the saturation pressure (" + str(satP) +
                   " bars) for this sample. Results from this calculation may be nonsensical.")

        # Use sympy to solve solubility equation for XH2Ofluid
        XCO2fluid = sympy.symbols('XCO2fluid')  # XCO2fluid is the variable to solve for

        equation = ((XCO2fluid*pressureMPa*(5668 - 55.99*(pressureMPa*(1-XCO2fluid)))/temperatureK
                    + (XCO2fluid*pressureMPa)*(0.4133*(pressureMPa*(1-XCO2fluid))**(0.5)
                    + 2.041*10**(-3)*(pressureMPa*(1-XCO2fluid))**(1.5))) - CO2melt_ppm)

        XCO2fluid = sympy.solve(equation, XCO2fluid, real=True)[0]

        if type(XCO2fluid) != float:
            w.warn("Could not find equilibrium fluid composition.")
            return 0
        else:
            if XCO2fluid > 1:
                XCO2fluid = 1
            if XCO2fluid < 0:
                XCO2fluid = 0

            return XCO2fluid

    def calculate_saturation_pressure(self, temperature, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a an CO2-bearing fluid is saturated. Calls the
        scipy.root_scalar routine, which makes repeated called to the
        calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample:        Sample class
            Magma major element composition.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 0. Mole fraction of CO2 in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        temperatureK = temperature + 273.15
        if temperatureK <= 0.0:
            raise core.InputError("Temperature must be greater than 0K.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0.0:
            raise core.InputError("Dissolved CO2 concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure,
                               args=(temperature, sample, X_fluid, kwargs),
                               x0=10.0, x1=2000.0).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return np.real(satP)

    def root_saturation_pressure(self, pressure, temperature, sample, X_fluid, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample:        Sample class
            Magma major element composition, including H2O.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, X_fluid=X_fluid, **kwargs) -
                sample.get_composition('CO2'))


# Defining compositional ranges for Liu - based on the Max value of the calibration dataset +-5%
# of that value
watercomprange = {'SiO2':  [71, 82],
                  'TiO2':  [0, 0.21],
                  'FeO':   [0, 1.5],
                  'Al2O3': [11.5, 14.2],
                  'MgO':   [0, 0.18],
                  'CaO':   [0, 1.2],
                  'Na2O':  [3.2, 4.9],
                  'K2O':   [3.4, 6.0]
                  }

carboncomprange = {'SiO2':  [73, 82],
                   'TiO2':  [0.07, 0.12],
                   'FeO':   [0.36, 1.1],
                   'Al2O3': [11.9, 13.7],
                   'MgO':   [0.05, 0.08],
                   'CaO':   [0.23, 0.6],
                   'Na2O':  [3.9, 4.4],
                   'K2O':   [4.0, 5.0]
                   }


def crf_WaterComp(calibval=None, sample=sample_class.Sample({})):
    comp = sample.get_composition(units='wtpt_oxides')

    test_results = []
    for ox in watercomprange.keys():
        test_results.append(comp[ox] >= watercomprange[ox][0] and
                            comp[ox] <= watercomprange[ox][1])

    return all(test_results)


def crf_CarbonComp(calibval=None, sample=sample_class.Sample({})):
    comp = sample.get_composition(units='wtpt_oxides')

    test_results = []
    for ox in carboncomprange.keys():
        test_results.append(comp[ox] >= carboncomprange[ox][0] and
                            comp[ox] <= carboncomprange[ox][1])

    return all(test_results)


def crf_MixedComp(calibval=None, sample=sample_class.Sample({})):
    comp = sample.get_composition(units='wtpt_oxides')

    test_results = []
    for ox in carboncomprange.keys():
        test_results.append(comp[ox] >= np.min([watercomprange[ox][0], carboncomprange[ox][0]]))
        test_results.append(comp[ox] <= np.max([watercomprange[ox][1], carboncomprange[ox][1]]))

    return all(test_results)


crmsg_Comp_pass = ("The sample appears to be similar in composition to the rhyolites and "
                   "haplogranites used to calibrate the Liu et al. model.")
crmsg_WaterComp_fail = (" These calibration limits were selected based on the minimum and "
                        "maximum values of these oxides (+-5%) in the Water calibration dataset. "
                        "As the Liu et al. model incorperates no term for compositional "
                        "dependence, users must take extreme care when extrapolating this model "
                        "to compositions which differ significantly from the haplogranites and "
                        "rhyolites in the calibration dataset. These warnings are simply a guide;"
                        " we suggest that users carefully compare their major element data to the "
                        "calibration dataset to check for suitability ")
crmsg_CarbonComp_fail = (" These calibration limits were selected based on the minimum and "
                         "maximum values of these oxides (+-5%) in the Carbon calibration dataset."
                         " As the Liu et al. model incorperates no term for compositional "
                         "dependence, users must take extreme care when extrapolating this model "
                         "to compositions which differ significantly from the haplogranites and "
                         "rhyolites in the calibration dataset. These warnings are simply a guide;"
                         " we suggest that users carefully compare their major element data to the"
                         " calibration dataset to check for suitability ")
crmsg_Comp_fail = (" These calibration limits were selected based on the minimum and maximum "
                   "values of these oxides (+-5%) in the combined Water and Carbon calibration"
                   " dataset. As the Liu et al. model incorperates no term for compositional "
                   "dependence, users must take extreme care when extrapolating this model to "
                   "compositions which differ significantly from the haplogranites and rhyolites "
                   "in the calibration dataset. These warnings are simply a guide; we suggest that"
                   " users carefully compare their major element data to the calibration dataset "
                   "to check for suitability ")
crmsg_Comp_description = "The Liu et al. model is suitable for haplogranites and rhyolites."


mixed = model_classes.MixedFluid({'H2O': water(), 'CO2': carbon()})

# Prevent the same error being returned for both H2O and CO2 when using the mixed model.
mixed.models[1].set_calibration_ranges([])
_crs_to_update = mixed.models[0].calibration_ranges
for _cr in _crs_to_update:
    _cr.model_name = 'Liu et al. (2005)'

cr_oxide_list = []
for ox in carboncomprange.keys():
    lo = np.min([watercomprange[ox][0], carboncomprange[ox][0]])
    hi = np.max([watercomprange[ox][1], carboncomprange[ox][1]])
    cr_oxide_list.append(
        calibration_checks.CalibrationRange(
            ox, [lo, hi], calibration_checks.crf_Between, 'wt%', 'Liu et al. (2005)',
            fail_msg=calibration_checks.crmsg_BC_fail,
            pass_msg=calibration_checks.crmsg_BC_pass,
            description_msg=calibration_checks.crmsg_Between_description))

mixed.models[0].set_calibration_ranges([
    calibration_checks.CalibrationRange(
        'pressure', [0, 5000.0], calibration_checks.crf_Between, 'bar', 'Liu et al. (2005)',
        fail_msg=calibration_checks.crmsg_Between_fail,
        pass_msg=calibration_checks.crmsg_Between_pass,
        description_msg=calibration_checks.crmsg_Between_description),
    calibration_checks.CalibrationRange(
        'temperature', [700.0, 1200], calibration_checks.crf_Between, 'oC', 'Liu et al. (2005)',
        fail_msg=calibration_checks.crmsg_Between_fail,
        pass_msg=calibration_checks.crmsg_Between_pass,
        description_msg=calibration_checks.crmsg_Between_description),
    calibration_checks.CalibrationRange(
        'sample', None, crf_MixedComp, None, None,
        fail_msg=crmsg_Comp_fail,
        pass_msg=crmsg_Comp_pass,
        description_msg=crmsg_Comp_description)] + cr_oxide_list)



from VESIcal.models import shishkina
from VESIcal.models import dixon
from VESIcal.models import iaconomarziano
from VESIcal.models import liu
from VESIcal.models import moore
from VESIcal.models import allison

default_models = {'ShishkinaIdealMixing':     shishkina.mixed,
                  'Dixon':                    dixon.mixed,
                  'IaconoMarziano':           iaconomarziano.mixed,
                  'Liu':                      liu.mixed,
                  'ShishkinaCarbon':          shishkina.carbon(),
                  'ShishkinaWater':           shishkina.water(),
                  'DixonCarbon':              dixon.carbon(),
                  'DixonWater':               dixon.water(),
                  'IaconoMarzianoCarbon':     iaconomarziano.carbon(),
                  'IaconoMarzianoWater':      iaconomarziano.water(),
                  'AllisonCarbon':            allison.vesuvius,
                  'AllisonCarbon_sunset':     allison.sunset,
                  'AllisonCarbon_sfvf':       allison.sfvf,
                  'AllisonCarbon_erebus':     allison.erebus,
                  'AllisonCarbon_vesuvius':   allison.vesuvius,
                  'AllisonCarbon_etna':       allison.etna,
                  'AllisonCarbon_stromboli':  allison.stromboli,
                  'MooreWater':               moore.water(),
                  'LiuWater':                 liu.water(),
                  'LiuCarbon':                liu.carbon()
                  }


def get_models(model='all'):
    """
    Returns model objects as a list

    Parameters
    ----------
    models:    str
        OPTIONAL. Default value is 'all' in which case all keys in default_models are returned.
        If 'mixed' is passed, only the MixedFluid model names are returned.
    """
    if model == 'all':
        return list(default_models.values())
    if model == 'mixed':
        # MagmaSat not included here as it is treated separately
        return [shishkina.mixed, dixon.mixed, iaconomarziano.mixed, liu.mixed]


def get_model_names(model='all'):
    """
    Returns all available model names as a list of strings.

    Parameters
    ----------
    models:    str
        OPTIONAL. Default value is 'all' in which case all keys in default_models are returned.
        If 'mixed' is passed, only the MixedFluid model names are returned.
    """
    if model == 'all':
        return list(default_models.keys())
    if model == 'mixed':
        # MagmaSat not included here as it is treated separately
        return ['ShishkinaIdealMixing', 'Dixon', 'IaconoMarziano', 'Liu']



from VESIcal import core
from VESIcal import models
from VESIcal import calculate_classes
from VESIcal import batchfile

from VESIcal.thermo import thermo_calculate_classes

import numpy as np
import warnings as w
import sys

from thermoengine import equilibrate

w.filterwarnings("ignore", message="rubicon.objc.ctypes_patch has only been "
                                   "tested ")

# -------------- MELTS preamble --------------- #
# instantiate thermoengine equilibrate MELTS instance
melts = equilibrate.MELTSmodel('1.2.0')

# Suppress phases not required in the melts simulation
phases = melts.get_phase_names()
for phase in phases:
    melts.set_phase_inclusion_status({phase: False})
melts.set_phase_inclusion_status({'Fluid': True, 'Liquid': True})
# --------------------------------------------- #


# -------------- BATCH PROCESSING ----------- #
class BatchFile(batchfile.BatchFile):
    """Performs model functions on a batchfile.BatchFile object
    """
    pass

    def get_XH2O_fluid(self, sample, temperature, pressure, H2O, CO2):
        """An internally used function to calculate fluid composition.

        Parameters
        ----------
        sample: dictionary
            Sample composition in wt% oxides

        temperature: float
            Temperature in degrees C.

        pressure: float
            Pressure in bars

        H2O: float
            wt% H2O in the system

        CO2: float
            wt% CO2 in the system

        Returns
        -------
        float
            Mole fraction of H2O in the H2O-CO2 fluid

        """
        pressureMPa = pressure / 10.0

        bulk_comp = {oxide:  sample[oxide] for oxide in core.magmasat_oxides}
        bulk_comp["H2O"] = H2O
        bulk_comp["CO2"] = CO2
        melts.set_bulk_composition(bulk_comp)

        output = melts.equilibrate_tp(temperature, pressureMPa,
                                      initialize=True)
        (status, temperature, pressureMPa, xmlout) = output[0]
        fluid_comp = melts.get_composition_of_phase(xmlout, phase_name='Fluid',
                                                    mode='component')
        # NOTE mode='component' returns endmember component keys with values
        # in mol fraction.

        if "Water" in fluid_comp:
            H2O_fl = fluid_comp["Water"]
        else:
            H2O_fl = 0.0

        return H2O_fl

    def calculate_dissolved_volatiles(self, temperature, pressure, X_fluid=1,
                                      print_status=True, model='MagmaSat',
                                      record_errors=False, **kwargs):
        """
        Calculates the amount of H2O and CO2 dissolved in a magma at the given
        P/T conditions and fluid composition. Fluid composition will be
        matched to within 0.0001 mole fraction.

        Parameters
        ----------
        temperature: float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case
            the passed value is used as the temperature for all samples.
            Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass
            the str value corresponding to the column title in the BatchFile
            object.

        pressure: float, int, or str
            Pressure, in bars. Can be passed as float or int, in which case the
            passed value is used as the pressure for all samples.
            Alternatively, pressure information for each individual sample may
            already be present in the BatchFile object. If so, pass the str
            value corresponding to the column title in the BatchFile object.

        X_fluid: float, int, or str
            OPTIONAL: Default value is 1. The mole fraction of H2O in the
            H2O-CO2 fluid. X_fluid=1 is a pure H2O fluid. X_fluid=0 is a pure
            CO2 fluid. Can be passed as a float or int, in which case the
            passed value is used as the X_fluid for all samples.
            Alternatively, X_fluid information for each individual sample may
            already be present in the BatchFile object. If so, pass the str
            value corresponding to the column title in the BatchFile object.

        print_status: bool
            OPTIONAL: The default value is True, in which case the progress of
            the calculation will be printed to the terminal. If set to False,
            nothing will be printed. MagmaSat calculations tend to be slow,
            and so a value of True is recommended for most use cases.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be
            passed here.

        record_errors: bool
            OPTIONAL: If True, any errors arising during the calculation will
            be recorded as a column.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        dissolved_data = self.get_data().copy()

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        if isinstance(pressure, str):
            file_has_press = True
            press_name = pressure
        elif isinstance(pressure, float) or isinstance(pressure, int):
            file_has_press = False
        else:
            raise core.InputError("pressure must be type str or float or int")

        if isinstance(X_fluid, str):
            file_has_X = True
            X_name = X_fluid
        elif isinstance(X_fluid, float) or isinstance(X_fluid, int):
            file_has_X = False
            if X_fluid != 0 and X_fluid != 1:
                if X_fluid < 0.001 or X_fluid > 0.999:
                    raise core.InputError("X_fluid is calculated to a "
                                          "precision of 0.0001 mole fraction. "
                                          "Value for X_fluid must be between "
                                          "0.0001 and 0.9999.")
        else:
            raise core.InputError("X_fluid must be type str or float or int")

        # Check if the model passed as the attribute "model_type"
        # Currently only implemented for MagmaSat type models
        if hasattr(model, 'model_type') is True:
            model = model.model_type

        H2Ovals = []
        CO2vals = []
        warnings = []
        errors = []
        if model in models.get_model_names(model='mixed'):
            for index, row in dissolved_data.iterrows():
                try:
                    if file_has_temp:
                        temperature = row[temp_name]

                    if file_has_press:
                        pressure = row[press_name]

                    if file_has_X:
                        X_fluid = row[X_name]

                    # Get sample comp as Sample class with defaults
                    bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                    bulk_comp.set_default_units(self.default_units)
                    bulk_comp.set_default_normalization(
                                                    self.default_normalization)
                    calc = calculate_classes.calculate_dissolved_volatiles(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           X_fluid=(X_fluid, 1-X_fluid),
                                           model=model, silence_warnings=True,
                                           **kwargs)
                    H2Ovals.append(calc.result['H2O_liq'])
                    CO2vals.append(calc.result['CO2_liq'])
                    warnings.append(calc.calib_check)
                    errors.append('')
                except Exception:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append('Calculation Failed.')
                    errors.append(sys.exc_info()[0])
            dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            dissolved_data["CO2_liq_VESIcal"] = CO2vals

            if file_has_temp is False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X is False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings
            if record_errors:
                dissolved_data["Errors"] = errors

            return dissolved_data

        elif model == 'MagmaSat':
            XH2Ovals = []
            XCO2vals = []
            FluidProportionvals = []
            iterno = 0
            for index, row in dissolved_data.iterrows():
                iterno += 1
                if print_status:
                    percent = iterno/len(dissolved_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp:
                    temperature = row[temp_name]
                if temperature <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad temperature.")
                    errors.append(sys.exc_info()[0])
                    w.warn("Temperature for sample " + str(index) +
                           " is <=0. Skipping sample.", stacklevel=2)

                if file_has_press:
                    pressure = row[press_name]
                if temperature > 0 and pressure <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad pressure.")
                    errors.append(sys.exc_info()[0])
                    w.warn("Pressure for sample " + str(index) +
                           " is <=0. Skipping sample.", stacklevel=2)

                if file_has_X:
                    X_fluid = row[X_name]
                if temperature > 0 and pressure > 0 and X_fluid < 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad X_fluid.")
                    errors.append(sys.exc_info()[0])
                    w.warn("X_fluid for sample " + str(index) +
                           " is <0. Skipping sample.", stacklevel=2)

                if temperature > 0 and pressure > 0 and X_fluid > 1:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    XH2Ovals.append(np.nan)
                    XCO2vals.append(np.nan)
                    FluidProportionvals.append(np.nan)
                    warnings.append("Sample skipped. Bad X_fluid.")
                    errors.append(sys.exc_info()[0])
                    w.warn("X_fluid for sample " + str(index) +
                           " is >1. Skipping sample.", stacklevel=2)

                if (temperature > 0 and pressure > 0 and
                   X_fluid >= 0 and X_fluid <= 1):
                    try:
                        # Get sample comp as Sample class with defaults
                        bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides',
                                      asSampleClass=True)
                        bulk_comp.set_default_units(self.default_units)
                        bulk_comp.set_default_normalization(
                                                    self.default_normalization)
                        calc = calculate_classes.calculate_dissolved_volatiles(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           X_fluid=X_fluid, model=model,
                                           silence_warnings=True, verbose=True)
                        H2Ovals.append(calc.result['H2O_liq'])
                        CO2vals.append(calc.result['CO2_liq'])
                        XH2Ovals.append(calc.result['XH2O_fl'])
                        XCO2vals.append(calc.result['XCO2_fl'])
                        FluidProportionvals.append(
                                             calc.result['FluidProportion_wt'])
                        warnings.append(calc.calib_check)
                        errors.append('')
                    except Exception:
                        H2Ovals.append(np.nan)
                        CO2vals.append(np.nan)
                        XH2Ovals.append(np.nan)
                        XCO2vals.append(np.nan)
                        FluidProportionvals.append(np.nan)
                        warnings.append('Calculation Failed.')
                        errors.append(sys.exc_info()[0])
            dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            dissolved_data["CO2_liq_VESIcal"] = CO2vals

            if file_has_temp is False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X is False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings
            if record_errors:
                dissolved_data["Errors"] = errors

            return dissolved_data

        else:
            XH2Ovals = []
            XCO2vals = []
            FluidProportionvals = []
            for index, row in dissolved_data.iterrows():
                if file_has_temp:
                    temperature = row[temp_name]
                if file_has_press:
                    pressure = row[press_name]
                if file_has_X:
                    X_fluid = row[X_name]
                # Get sample comp as Sample class with defaults
                bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides',
                                      asSampleClass=True)
                bulk_comp.set_default_units(self.default_units)
                bulk_comp.set_default_normalization(self.default_normalization)
                if 'Water' in model:
                    try:
                        calc = calculate_classes.calculate_dissolved_volatiles(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           X_fluid=X_fluid, model=model,
                                           silence_warnings=True)
                        H2Ovals.append(calc.result)
                        warnings.append(calc.calib_check)
                    except Exception:
                        H2Ovals.append(0)
                        warnings.append('Calculation Failed #001')
                if 'Carbon' in model:
                    try:
                        calc = calculate_classes.calculate_dissolved_volatiles(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           X_fluid=X_fluid, model=model,
                                           silence_warnings=True)
                        CO2vals.append(calc.result)
                        warnings.append(calc.calib_check)
                    except Exception:
                        CO2vals.append(0)
                        warnings.append('Calculation Failed #002')

            if 'Water' in model:
                dissolved_data["H2O_liq_VESIcal"] = H2Ovals
            if 'Carbon' in model:
                dissolved_data["CO2_liq_VESIcal"] = CO2vals
            if file_has_temp is False:
                dissolved_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                dissolved_data["Pressure_bars_VESIcal"] = pressure
            if file_has_X is False:
                dissolved_data["X_fluid_input_VESIcal"] = X_fluid
            dissolved_data["Model"] = model
            dissolved_data["Warnings"] = warnings

            return dissolved_data

    def calculate_equilibrium_fluid_comp(self, temperature, pressure=None,
                                         print_status=False, model='MagmaSat',
                                         **kwargs):
        """
        Returns H2O and CO2 concentrations in wt% or mole fraction in a fluid
        in equilibrium with the given sample(s) at the given P/T condition.

        Parameters
        ----------
        sample: BatchFile object
            Compositional information on samples in oxides.

        temperature: float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case
            the passed value is used as the temperature for all samples.
            Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass
            the str value corresponding to the column title in the  BatchFile
            object.

        presure: float, int, or str
            Pressure, in bars. Can be passed as float or int, in which case
            the passed value is used as the pressure for all samples.
            Alternatively, pressure information for each individual sample may
            already be present in the BatchFile object. If so, pass the str
            value corresponding to the column title in the BatchFile object.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be
            passed here.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        fluid_data = self.get_data().copy()

        # Check if the model passed as the attribute "model_type"
        # Currently only implemented for MagmaSat type models
        if hasattr(model, 'model_type') is True:
            model = model.model_type

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        if isinstance(pressure, str):
            file_has_press = True
            press_name = pressure
        elif (isinstance(pressure, float) or isinstance(pressure, int) or
              pressure is None):
            file_has_press = False
        else:
            raise core.InputError("pressure must be type str or float or int")

        H2Ovals = []
        CO2vals = []
        warnings = []
        if kwargs.get('verbose') is True:
            FluidMass_grams_vals = []
            FluidProportion_wt_vals = []
        if (model in models.get_model_names(model='mixed') or
           model == "MooreWater"):
            for index, row in fluid_data.iterrows():
                try:
                    if file_has_temp:
                        temperature = row[temp_name]
                    if file_has_press:
                        pressure = row[press_name]
                    # Get sample comp as Sample class with defaults
                    bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                    bulk_comp.set_default_units(self.default_units)
                    bulk_comp.set_default_normalization(
                                                    self.default_normalization)

                    calc = calculate_classes.calculate_equilibrium_fluid_comp(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           model=model, silence_warnings=True,
                                           **kwargs)

                    H2Ovals.append(calc.result['H2O'])
                    CO2vals.append(calc.result['CO2'])
                    if calc.result['H2O'] == 0 and calc.result['CO2'] == 0:
                        warnings.append(calc.calib_check + "Sample not " +
                                        "saturated at these conditions")
                    else:
                        warnings.append(calc.calib_check)
                except Exception:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation Failed.")
            fluid_data["XH2O_fl_VESIcal"] = H2Ovals
            fluid_data["XCO2_fl_VESIcal"] = CO2vals
            if file_has_temp is False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data
        elif model == 'MagmaSat':
            iterno = 0
            for index, row in fluid_data.iterrows():
                iterno += 1
                if print_status:
                    percent = iterno/len(fluid_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp:
                    temperature = row[temp_name]
                if temperature <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation skipped. Bad temperature.")
                    w.warn("Temperature for sample " + str(index) +
                           " is <=0. Skipping sample.",
                           stacklevel=2)

                if file_has_press:
                    pressure = row[press_name]
                if temperature > 0 and pressure <= 0:
                    H2Ovals.append(np.nan)
                    CO2vals.append(np.nan)
                    warnings.append("Calculation skipped. Bad pressure.")
                    w.warn("Pressure for sample " + str(index) +
                           " is <=0. Skipping sample.", stacklevel=2)

                if temperature > 0 and pressure > 0:
                    try:
                        # Get sample comp as Sample class with defaults
                        bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                        bulk_comp.set_default_units(self.default_units)
                        bulk_comp.set_default_normalization(
                                                    self.default_normalization)

                        calc = (
                            calculate_classes.calculate_equilibrium_fluid_comp(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           model=model, silence_warnings=True,
                                           **kwargs))

                        H2Ovals.append(calc.result['H2O'])
                        CO2vals.append(calc.result['CO2'])
                        if kwargs.get('verbose') is True:
                            FluidMass_grams_vals.append(calc.result['FluidMass_grams'])
                            FluidProportion_wt_vals.append(calc.result['FluidProportion_wt'])
                        if calc.result['H2O'] == 0 and calc.result['CO2'] == 0:
                            warnings.append(calc.calib_check + "Sample not " +
                                            "saturated at these conditions")
                        else:
                            warnings.append(calc.calib_check)
                    except Exception:
                        H2Ovals.append(np.nan)
                        CO2vals.append(np.nan)
                        warnings.append("Calculation Failed.")
                        if kwargs.get('verbose') is True:
                            FluidMass_grams_vals.append(np.nan)
                            FluidProportion_wt_vals.append(np.nan)
            fluid_data["XH2O_fl_VESIcal"] = H2Ovals
            fluid_data["XCO2_fl_VESIcal"] = CO2vals
            if kwargs.get('verbose') is True:
                fluid_data["FluidMass_grams"] = FluidMass_grams_vals
                fluid_data["FluidProportion_wt"] = FluidProportion_wt_vals
            if file_has_temp is False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data

        else:
            saturated = []
            for index, row in fluid_data.iterrows():
                try:
                    if file_has_temp:
                        temperature = row[temp_name]
                    if file_has_press:
                        pressure = row[press_name]
                    # Get sample comp as Sample class with defaults
                    bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                    bulk_comp.set_default_units(self.default_units)
                    bulk_comp.set_default_normalization(
                                                    self.default_normalization)

                    calc = calculate_classes.calculate_equilibrium_fluid_comp(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature,
                                           model=model, silence_warnings=True)
                    saturated.append(calc.result)
                    warnings.append(calc.calib_check)
                except Exception:
                    saturated.append(np.nan)
                    warnings.append("Calculation Failed.")
            fluid_data["Saturated_VESIcal"] = saturated
            if file_has_temp is False:
                fluid_data["Temperature_C_VESIcal"] = temperature
            if file_has_press is False:
                fluid_data["Pressure_bars_VESIcal"] = pressure
            fluid_data["Model"] = model
            fluid_data["Warnings"] = warnings

            return fluid_data

    def calculate_saturation_pressure(self, temperature, print_status=None,
                                      model='MagmaSat', **kwargs):
        """
        Calculates the saturation pressure of multiple sample compositions in
        the BatchFile.

        Parameters
        ----------
        temperature: float, int, or str
            Temperature at which to calculate saturation pressures, in
            degrees C. Can be passed as float or int, in which case the
            passed value is used as the temperature for all samples.
            Alternatively, temperature information for each individual sample
            may already be present in the passed BatchFile object. If so, pass
            the str value corresponding to the column title in the passed
            BatchFile object.

        print_status: bool
            OPTIONAL: The default value for MagmaSat is True and the default
            for all other models is False. If set to True, the progress of the
            calculation will be printed to the terminal. If set to False,
            nothing will be printed. MagmaSat calculations tend to be slow,
            and so a value of True is recommended more most use cases.

        model: string
            OPTIONAL: Default is 'MagmaSat'. Any other model name can be
            passed here.

        Returns
        -------
        pandas DataFrame object
            Values returned are saturation pressure in bars, the mass of
            fluid present, and the composition of the fluid present.
        """
        satp_data = self.get_data().copy()

        # Check if the model passed has the attribute "model_type"
        # Currently only implemented for MagmaSat type models
        if hasattr(model, 'model_type') is True:
            model = model.model_type

        # set default print_status to True for MagmaSat, False for other
        # models if user doesn't pass any option
        if print_status is None:
            if model == 'MagmaSat':
                print_status = True
            else:
                print_status = False

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temperature must be type str or float or "
                                  "int")

        if model != 'MagmaSat':
            satP = []
            warnings = []
            piStar = []
            iterno = 0
            for index, row in satp_data.iterrows():
                iterno += 1
                if print_status:
                    percent = iterno/len(satp_data.index)
                    batchfile.status_bar.status_bar(percent, index)
                if file_has_temp:
                    temperature = row[temp_name]
                # Get sample comp as Sample class with defaults
                bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                bulk_comp.set_default_units(self.default_units)
                bulk_comp.set_default_normalization(self.default_normalization)

                calc = calculate_classes.calculate_saturation_pressure(
                                     sample=bulk_comp, temperature=temperature,
                                     model=model, silence_warnings=True,
                                     **kwargs)
                satP.append(calc.result)
                warnings.append(calc.calib_check)
                if model == 'ShishkinaIdealMixing':
                    piStar.append(
                           models.default_models['ShishkinaIdealMixing'
                                                 ].models[1].PiStar(bulk_comp))

            satp_data["SaturationP_bars_VESIcal"] = satP
            if file_has_temp is False:
                satp_data["Temperature_C_VESIcal"] = temperature
            satp_data["Model"] = model
            satp_data["Warnings"] = warnings
            if len(piStar) > 0:
                satp_data['PiStar_VESIcal'] = piStar

            return satp_data

        elif model == 'MagmaSat':
            satP = []
            flmass = []
            flH2O = []
            flCO2 = []
            flsystem_wtper = []
            warnings = []
            iterno = 0
            for index, row in satp_data.iterrows():
                iterno += 1
                if print_status:
                    percent = iterno/len(satp_data.index)
                    batchfile.status_bar.status_bar(percent, index)

                if file_has_temp:
                    temperature = row[temp_name]
                if temperature <= 0:
                    satP.append(np.nan)
                    flmass.append(np.nan)
                    flsystem_wtper.append(np.nan)
                    flH2O.append(np.nan)
                    flCO2.append(np.nan)
                    warnings.append("Calculation skipped. Bad temperature.")
                    w.warn("Temperature for sample " + str(index) +
                           " is <=0. Skipping sample.", stacklevel=2)

                if temperature > 0:
                    try:
                        # Get sample comp as Sample class with defaults
                        bulk_comp = self.get_sample_composition(
                                      index,
                                      normalization=self.default_normalization,
                                      units='wtpt_oxides', asSampleClass=True)
                        bulk_comp.set_default_units(self.default_units)
                        bulk_comp.set_default_normalization(
                                                    self.default_normalization)

                        calc = calculate_classes.calculate_saturation_pressure(
                                     sample=bulk_comp, temperature=temperature,
                                     model=model, verbose=True,
                                     silence_warnings=True)
                        satP.append(calc.result["SaturationP_bars"])
                        flmass.append(calc.result["FluidMass_grams"])
                        flsystem_wtper.append(
                                             calc.result["FluidProportion_wt"])
                        flH2O.append(calc.result["XH2O_fl"])
                        flCO2.append(calc.result["XCO2_fl"])
                        warnings.append(calc.calib_check)
                    except Exception:
                        satP.append(np.nan)
                        flmass.append(np.nan)
                        flsystem_wtper.append(np.nan)
                        flH2O.append(np.nan)
                        flCO2.append(np.nan)
                        warnings.append("Calculation Failed")

            satp_data["SaturationP_bars_VESIcal"] = satP
            if file_has_temp is False:
                satp_data["Temperature_C_VESIcal"] = temperature
            satp_data["XH2O_fl_VESIcal"] = flH2O
            satp_data["XCO2_fl_VESIcal"] = flCO2
            satp_data["FluidMass_grams_VESIcal"] = flmass
            satp_data["FluidSystem_wt_VESIcal"] = flsystem_wtper
            satp_data["Model"] = model
            satp_data["Warnings"] = warnings

            return satp_data

    def calculate_liquid_density(self, temperature, pressure,
                                 record_errors=False, **kwargs):
        """ Calculates the density of the liquid using the DensityX model. Using
        this interface will preprocess the sample, run the calculation, and then
        check the calibration ranges. All parameters required by the chosen
        model must be passed.

        Parameters
        ----------
        pressure:   float, int, or str
            Pressure, in bars. Can be passed as float or int, in which case the
            passed value is used as the pressure for all samples.
            Alternatively, pressure information for each individual sample may
            already be present in the BatchFile object. If so, pass the str
            value corresponding to the column title in the BatchFile object.

        temperature:  float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case
            the passed value is used as the temperature for all samples.
            Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass
            the str value corresponding to the column title in the BatchFile
            object.

        record_errors: bool
            OPTIONAL: If True, any errors arising during the calculation will
            be recorded as a column.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        density_data = self.get_data().copy()

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        if isinstance(pressure, str):
            file_has_press = True
            press_name = pressure
        elif isinstance(pressure, float) or isinstance(pressure, int):
            file_has_press = False
        else:
            raise core.InputError("pressure must be type str or float or int")

        density_vals = []
        warnings = []
        errors = []
        for index, row in density_data.iterrows():
            try:
                if file_has_temp:
                    temperature = row[temp_name]

                if file_has_press:
                    pressure = row[press_name]

                # Get sample comp as Sample class with defaults
                bulk_comp = self.get_sample_composition(index,
                                                        normalization=self.default_normalization,
                                                        units='wtpt_oxides', asSampleClass=True)
                bulk_comp.set_default_units(self.default_units)
                bulk_comp.set_default_normalization(self.default_normalization)
                calc = thermo_calculate_classes.calculate_liquid_density(
                                           sample=bulk_comp, pressure=pressure,
                                           temperature=temperature, **kwargs)
                density_vals.append(calc.result)
                warnings.append(calc.calib_check)
                errors.append('')
            except Exception:
                density_vals.append(np.nan)
                warnings.append('Calculation Failed')
                errors.append(sys.exc_info()[0])
        density_data["Density_liq_VESIcal"] = density_vals

        if file_has_temp is False:
            density_data["Temperature_C_VESIcal"] = temperature
        if file_has_press is False:
            density_data["Pressure_bars_VESIcal"] = pressure
        density_data["Model"] = "DensityX"
        density_data["Warnings"] = warnings

        if record_errors:
            density_data["Errors"] = errors

        return density_data

    def calculate_liquid_viscosity(self, temperature, record_errors=False,
                                   **kwargs):
        """
        Calculates the viscosity of the liquid using the Giordano et al.
        (2008) model. Using this interface will preprocess the sample, run the
        calculation, and then check the calibration ranges. All parameters
        required by the chosen model must be passed.

        Parameters
        ----------
        temperature:  float, int, or str
            Temperature, in degrees C. Can be passed as float, in which case
            the passed value is used as the temperature for all samples.
            Alternatively, temperature information for each individual
            sample may already be present in the BatchFile object. If so, pass
            the str value corresponding to the column title in the BatchFile
            object.

        record_errors: bool
            OPTIONAL: If True, any errors arising during the calculation will
            be recorded as a column.

        Returns
        -------
        pandas DataFrame
            Original data passed plus newly calculated values are returned.
        """
        viscosity_data = self.get_data().copy()

        if isinstance(temperature, str):
            file_has_temp = True
            temp_name = temperature
        elif isinstance(temperature, float) or isinstance(temperature, int):
            file_has_temp = False
        else:
            raise core.InputError("temp must be type str or float or int")

        viscosity_vals = []
        warnings = []
        errors = []
        for index, row in viscosity_data.iterrows():
            try:
                if file_has_temp:
                    temperature = row[temp_name]

                # Get sample comp as Sample class with defaults
                bulk_comp = self.get_sample_composition(index,
                                                        normalization=self.default_normalization,
                                                        units='wtpt_oxides', asSampleClass=True)
                bulk_comp.set_default_units(self.default_units)
                bulk_comp.set_default_normalization(self.default_normalization)
                calc = thermo_calculate_classes.calculate_liquid_viscosity(
                                           sample=bulk_comp,
                                           temperature=temperature, **kwargs)
                viscosity_vals.append(calc.result)
                warnings.append(calc.calib_check)
                errors.append('')
            except Exception:
                viscosity_vals.append(np.nan)
                warnings.append('Calculation Failed')
                errors.append(sys.exc_info()[0])
        viscosity_data["Viscosity_liq_VESIcal"] = viscosity_vals

        if file_has_temp is False:
            viscosity_data["Temperature_C_VESIcal"] = temperature
        viscosity_data["Model"] = "Giordano et al. (2008)"
        viscosity_data["Warnings"] = warnings

        if record_errors:
            viscosity_data["Errors"] = errors

        return viscosity_data


def BatchFile_from_DataFrame(dataframe, units='wtpt_oxides', label=None):
    """
    Transforms any pandas DataFrame object into a VESIcal BatchFile object.

    Parameters
    ----------
    Same as batchfile.BatchFile_from_DataFrame()

    Returns
    -------
    VESIcal.BatchFile object
    """
    return BatchFile(filename=None, dataframe=dataframe, units=units,
                     label=label)



"""
VESIcal

A generalized python library for calculating and plotting various things
related to mixed volatile (H2O-CO2) solubility in silicate melts.
"""

__version__ = "1.1.1"
__author__ = "Kayla Iacovino, Simon Matthews, and Penny Wieser"

# ----------------- IMPORTS ----------------- #
import warnings as w
import pandas as pd

import VESIcal.core
from VESIcal.core import oxides, anhydrous_oxides, volatiles  # noqa F401
from VESIcal.core import fluid_molfrac_to_wt, fluid_wt_to_molfrac  # noqa F401
import VESIcal.activity_models
import VESIcal.batchfile
import VESIcal.batchmodel
import VESIcal.calculate_classes
import VESIcal.calibration_checks
import VESIcal.calibrations
import VESIcal.fugacity_models
import VESIcal.models
import VESIcal.sample_class
import VESIcal.vplot
import VESIcal.thermo

# -------------- TURN OFF WARNINGS ------------- #
w.filterwarnings("ignore", message="rubicon.objc.ctypes_patch has only been tested ")
w.filterwarnings("ignore", message="The handle")


# -------------- CALCULATION DEFINITIONS ----- #
class calculate_dissolved_volatiles(
    VESIcal.calculate_classes.calculate_dissolved_volatiles
):
    pass


class calculate_equilibrium_fluid_comp(
    VESIcal.calculate_classes.calculate_equilibrium_fluid_comp
):
    pass


class calculate_isobars_and_isopleths(
    VESIcal.calculate_classes.calculate_isobars_and_isopleths
):
    pass


class calculate_saturation_pressure(
    VESIcal.calculate_classes.calculate_saturation_pressure
):
    pass


class calculate_degassing_path(VESIcal.calculate_classes.calculate_degassing_path):
    pass


class calculate_liquid_density(
    VESIcal.thermo.thermo_calculate_classes.calculate_liquid_density
):
    pass


class calculate_liquid_viscosity(
    VESIcal.thermo.thermo_calculate_classes.calculate_liquid_viscosity
):
    pass


# -------------- ACCESS TO GET_MODEL_NAMES ----- #
def get_model_names(model="all"):
    return VESIcal.models.get_model_names(model=model)


# -------------- PLOTTING DEFINITIONS ----- #
def plot(**kwargs):
    """
    Custom automatic plotting of model calculations in VESIcal.
    Isobars, isopleths, and degassing paths can be plotted. Labels can be
    specified for each. Any combination of isobars, isopleths, and degassing
    paths can be plotted.

    Parameters
    ----------
    isobars: pandas DataFrame or list
        OPTIONAL. DataFrame object containing isobar information as calculated
        by calculate_isobars_and_isopleths. Or a list of DataFrame objects.

    isopleths: pandas DataFrame or list
        OPTIONAL. DataFrame object containing isopleth information as
        calculated by calculate_isobars_and_isopleths. Or a list of DataFrame
        objects.

    degassing_paths: list
        OPTIONAL. List of DataFrames with degassing information as generated
        by calculate_degassing_path().

    custom_H2O: list
        OPTIONAL. List of groups of H2O values to plot as points. For example
        myfile.data['H2O'] is one group of H2O values. Must be passed with
        custom_CO2 and must be same length as custom_CO2.

    custom_CO2: list
        OPTIONAL. List of groups of CO2 values to plot as points.For example
        myfile.data['CO2'] is one group of CO2 values. Must be passed with
        custom_H2O and must be same length as custom_H2O.

    isobar_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted line will be given the generic legend name of
        "Isobars n", with n referring to the nth isobars passed. Isobar
        pressure is given in parentheses. The user can pass their own labels
        as a list of strings. If more than one set of isobars is passed, the
        labels should refer to each set of isobars, not each pressure.

    isopleth_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted isopleth will be given the generic legend name of
        "Isopleth n", with n referring to the nth isopleths passed. Isopleth
        XH2O values are given in parentheses. The user can pass their own
        labels as a list of strings. If more than one set of isopleths is
        passed, the labels should refer to each set of isopleths, not each
        XH2O value.

    degassing_path_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each plotted line will be given the generic legend name of "Pathn",
        with n referring to the nth degassing path passed. The user can pass
        their own labels as a list of strings.

    custom_labels: list
        OPTIONAL. Labels for the plot legend. Default is None, in which case
        each group of custom points will be given the generic legend name of
        "Customn", with n referring to the nth degassing path passed. The user
        can pass their own labels as a list of strings.

    custom_colors: list
        OPTIONAL. Default value is "VESIcal", which uses VESIcal's color ramp.
        A list of color values readable by matplotlib can be passed here if
        custom symbol colors are desired. The length of this list must match
        that of custom_H2O and custom_CO2.

    custom_symbols: list
        OPTIONAL. Default value is None, in which case data are plotted as
        filled circles.. A list of symbol tyles readable by matplotlib can be
        passed here if custom symbol types are desired. The length of this
        list must match that of custom_H2O and custom_CO2.

    markersize: int
        OPTIONAL. Default value is 10. Same as markersize kwarg in matplotlib.
        Any numeric value passed here will set the marker size for
        (custom_H2O, custom_CO2) points.

    figsize: tuple
        OPTIONAL. Default value is (12,8). Sets the matplotlib.pyplot figsize
        value as (x_dimension, y_dimension)

    save_fig: False or str
        OPTIONAL. Default value is False, in which case the figure will not be
        saved. If a string is passed, the figure will be saved with the string
        as the filename. The string must include the file extension.

    extend_isobars_to_zero: bool
        OPTIONAL. If True (default), isobars will be extended to zero, even if
        there is a finite solubility at zero partial pressure.

    smooth_isobars: bool
        OPTIONAL. Default is False. If set to True, isobar data will be fit to
        a polynomial and plotted. If False, the raw input data will be plotted.

    smooth_isopleths: bool
        OPTIONAL. Default is False. If set to True, isopleth data will be fit
        to a polynomial and plotted. If False, the raw input data will be
        plotted.

    Returns
    -------
    fig, axes Matplotlib objects
        fig and axes matploblib objects defining a plot with x-axis as H2O wt%
        in the melt and y-axis as CO2 wt%in the melt. Isobars, or lines of
        constant pressure at which the sample magma composition is saturated,
        and isopleths, or lines of constant fluid composition at which the
        sample magma composition is saturated, are plotted if passed.
        Degassing paths, or the concentration of dissolved H2O and CO2 in a
        melt equilibrated along a path of decreasing pressure, is plotted if
        passed.
    """
    return VESIcal.vplot.plot(**kwargs)


def calib_plot(**kwargs):
    """
    Plots user data and calibration set of any or all models on any x-y plot
    or a total alkalis vs silica (TAS) diagram. TAS diagram boundaries
    provided by tasplot python module, copyright John A Stevenson.

    Parameters
    ----------
    user_data: BatchFile object, Sample object, pandas DataFrame, pandas Series,
        or dict.
        OPTIONAL. Default value is None, in which case only the model
        calibration set is plotted. User provided sample data describing the
        oxide composition of one or more samples. Multiple samples can be
        passed as an BatchFile object or pandas DataFrame. A single sample can
        be passed as a pandas Series.

    model: str or list
        OPTIONAL. Default value is 'all', in which case all model calibration
        datasets will be plotted. 'Mixed' can be used to plot all mixed fluid
        models. String of the name of the model calibration dataset to plot
        (e.g., 'Shishkina'). Multiple models can be plotted by passing them as
        strings within a list (e.g., ['Shishkina', 'Dixon']).

    plot_type: str
        OPTIONAL. Default value is 'TAS', which returns a total alkali vs
        silica (TAS) diagram. Any two oxides can be plotted as an x-y plot by
        setting plot_type='xy' and specifying x- and y-axis oxides, e.g.,
        x='SiO2', y='Al2O3'.

    zoom: str or list
        OPTIONAL. Default value is None in which case axes will be set to the
        default of 35<x<100 wt% and 0<y<25 wt% for TAS type plots and the best
        values to show the data for xy type plots. Can pass "user_data" to
        plot the figure where the x and y axes are scaled down to zoom in and
        only show the region surrounding the user_data. A list of tuples may
        be passed to manually specify x and y limits. Pass in data as
        [(x_min, x_max), (y_min, y_max)]. For example, the default limits here
        would be passed in as [(35,100), (0,25)].

    figsize: tuple
        OPTIONAL. Default value is (17,8). Sets the matplotlib.pyplot figsize
        value as (x_dimension, y_dimension).

    legend: bool
        OPTIONAL. Default value is True. Can be set to False in which case the
        legend will not be displayed.

    save_fig: False or str
        OPTIONAL. Default value is False, in which case the figure will not be
        saved. If a string is passed, the figure will be saved with the string
        as the filename. The string must include the file extension.

    Returns
    -------
    matplotlib object
    """
    return VESIcal.vplot.calib_plot(**kwargs)


def show():
    """
    Inherits from vplot.show()
    """
    return VESIcal.vplot.show()


# -------------- SAMPLE PROCESSING ---------- #
class Sample(VESIcal.sample_class.Sample):
    """The sample class stores compositional information for samples, and contains methods for
    normalization and other compositional calculations.

    The composition is stored as wtpt. If the composition is provided as wtpt, no
    normalization will be applied. If the composition is supplied as mols, the composition
    will be normalized to 100 wt%.

    Parameters
    ----------
    composition     dict or pandas.Series
        The composition of the sample in the format specified by the composition_type
        parameter. Default is oxides in wtpt.

    units     str
        Specifies the units and type of compositional information passed in the composition
        parameter. Choose from 'wtpt_oxides', 'mol_oxides', 'mol_cations'.

    default_normalization:     None or str
        The type of normalization to apply to the data by default. One of:
            - None (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including
              volatiles. The volatile wt% will remain fixed, whilst the other major element
              oxides are reduced proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it
              is volatile-free. If H2O or CO2 are passed to the function, their un-normalized
              values will be retained in addition to the normalized non-volatile oxides,
              summing to >100%.

    default_units     str
        The type of composition to return by default, one of:
        - wtpt_oxides (default)
        - mol_oxides
        - mol_cations
        - mol_singleO
    """

    pass


def get_oxides(sample):
    """
    Returns a sample composition with only compositional oxide data, removing
    any extranneous data. Useful when passing a self-defined sample (e.g. dict
    or pandas Series) to a some VESIcal function.

    Parameters
    ----------
    sample: pandas Series or dictionary
    A sample composition plus other sample information

    Returns
    -------
    Same type as passed sample (pandas Series or dictionary)
    Sample composition with extranneous information removed.
    """

    clean = {oxide: sample[oxide] for oxide in VESIcal.core.oxides}

    if isinstance(sample, dict):
        return clean
    if isinstance(sample, pd.core.series.Series):
        return pd.Series(clean)


# -------------- BATCH PROCESSING ------------ #
class BatchFile(VESIcal.batchmodel.BatchFile):
    """A batch file with sample names and oxide compositions

    Attributes
    ----------
        filename: str
            Path to the batch file, e.g., "my_file.xlsx". This always needs to
            be passed, even if the user is passing a pandas DataFrame rather
            than an batch file. If passing a DataFrame, filename should be set
            to None. File can be excel file (.xlsx) or .csv.

        sheet_name: str
            OPTIONAL. For Excel files. Default value is 0 which gets the first
            sheet in the batch spreadsheet file. This implements the pandas.
            read_excel() sheet_name parameter. But functionality to read in
            more than one sheet at a time (e.g.,
            pandas.read_excel(sheet_name=None)) is not yet imlpemented in
            VESIcal. From the pandas 1.0.4 documentation:

            Available cases:
            - Defaults to 0: 1st sheet as a DataFrame
            - 1: 2nd sheet as a DataFrame
            - "Sheet1": Load sheet with name “Sheet1”

        file_type: str
            OPTIONAL. Default is 'excel', which denotes that passed file has
            extension .xlsx. Other option is 'csv', which denotes that the
            passed file has extension .csv.

        units: str
            OPTIONAL. Default is 'wtpt_oxides'. String defining whether the
            oxide composition is given in wt percent ("wtpt_oxides", which is
            the default), mole oxides (mol_oxides) or mole cations
            (mol_cations).

        default_normalization:     None or str
            The type of normalization to apply to the data by default. One of:
            - None (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%,
               including volatiles. The volatile wt% will remain fixed, whilst
               the other major element oxides are reduced proportionally so
               that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to
              100%, assuming it is volatile-free. If H2O or CO2 are passed to
              the function, their un-normalized values will be retained in
              addition to the normalized non-volatile oxides, summing to >100%.

        default_units     str
            The type of composition to return by default, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        label: str
            OPTIONAL. Default is 'Label'. Name of the column within the passed
            file referring to sample names.

        dataframe: pandas DataFrame
            OPTIONAL. Default is None in which case this argument is ignored.
            This argument is used when the user wishes to turn a pandas
            DataFrame into an BatchFile object, for example when user data is
            already in python rather than being imported from a file. In this
            case set `dataframe` equal to the dataframe object being passed in.
            If using this option, pass None to filename.
    """

    pass


def BatchFile_from_DataFrame(dataframe, **kwargs):
    """
    Provides method for creating a BatchFile object from an existing pandas
    DataFrame. Inherits from batchfile.BatchFile().
    """
    return VESIcal.batchmodel.BatchFile_from_DataFrame(dataframe, **kwargs)


"""
                        ,,,                                     .*****
                       ,***,*                                  ,* *****
                      ,***,,,,*                              ,* ,*******
                     .****,,,.,**                          ,,*.,*********
                     *****,,,,.****                     ,,,,*.,********* *
                     *,**,,,,,,,******                ,,,,,* ,,********,.*                                     .                # noqa: E501
                    .*,*,, //,,.,*******            ,,,,,**.//,*******,,.,                                   //                 # noqa: E501
                    ,**,, //// ,,******** ,,,,,,,, **,,,*** ((,,,,,,,,,,,*                                  //*####             # noqa: E501
                    .,* , ///((,,,******. ,,,,,,, *********(((,.#### ,,.,*                                   ///#######         # noqa: E501
                     ,,** .((((,******** ,,,*,,, *********.(((#####/.,,**    .,,                              ///#*#######      # noqa: E501
                     ,,**(((((.******* ,,**** **.*********.#(####(( .,**.******                                ///##/#######    # noqa: E501
                      ,,***(.( ****** ,*******#*********** #*###( ,,** ******,                              /////(###########   # noqa: E501
                    ,,, ,**,,,,,,. ,,,**,**.### ****.************.  .******,                        .,,,,,****, ##############  # noqa: E501
                    .,,,,,,.,,,,,,,,,* *,#### ************# *************,                    ,,,,********,***/#######(#######  # noqa: E501
                      ,,,,,, /(   (#,*##### *****###/ #### ************                 ,,,,,,**********,.**.#######(#########( # noqa: E501
                        ,,,,, ((((((# ### * * #..######## *****************.##        ,,,,,****************,####. (## #######   # noqa: E501
                      ,,,,,,,, *(((( *###.****,#  ##### *********** #### ###         ,,,,,*********************,######*#######  # noqa: E501
                     ./ ,,,,,,,   ,****########*****,       .**########### ,         ,,,*********************** (##./##.##/#*   # noqa: E501
                       /////*,  ,,  *****,*,******   ,,.,((( ########### ,****      .,,.*****************,######  //#/////,     # noqa: E501
                        //////  /(,. *****###.****  *((,*&((########## *****       .****,************ ###########(/////*  //    # noqa: E501
                          /////,*#((/,(((*########* *((#(*##########********** .,   *****,*******    .###########//////,        # noqa: E501
                             ///// ,* (((((####### %#  ##########,**********,************* **  **** /##### ///////*/ /          # noqa: E501
                                ./////((((((################# ###***************************** ,,,,,  ///////*  ////            # noqa: E501
                              ///// .////(((############ .(((###**#.******,********************     .    //////.                # noqa: E501
                                * //// ////(######## ///(#########*******,****.*****************                                # noqa: E501
                                 *//////.////### */////(#########.*****,,****** ****************                                # noqa: E501
                               /////////////////////##/##########* .* ,,*******,,************ **.                               # noqa: E501
                                   ,///////////////################(,,,********,*  ************ *,*                             # noqa: E501
                                        , ////////################,,,,*********  ..,******,** **.***                            # noqa: E501
                                        , //////##/############# ,,,,,******* *****.,,*.,,, *********,,                         # noqa: E501
                                       ,,,//////(#/############  ,,,,,************ ((( ,, .***********                          # noqa: E501
                                       ,,///  .  ///###########((,,,,,*********.* //   ,  ,** *  *,                             # noqa: E501
                                       ,,,,,,,,,,  //#########,(( ,,,,******** *       ,, ,,**   *,                             # noqa: E501
                                       ,,,,,,,,,.   *//###( #,/// ,,,,******* *        .,,, ,,***..                             # noqa: E501
                                       .,,,***,,      ///# *,//*,,.,,,****** ,           ,,,,,,.,,,,                            # noqa: E501
                                       ,,,*****         *  // ,,,. ,,,***** *              ,,,,,,,,*.                           # noqa: E501
                                       , ,*****              ,, ,, *,,******                 ,,,,,,**                           # noqa: E501
                                       ,,,.***               ,,,,, ** *.****                  , ,,,,,                           # noqa: E501
                                       ,,,****              ,,,,,  ********                    ,,,,,,                           # noqa: E501
                                       ,,****               ,,,,,  *******                      ,,,,,                           # noqa: E501
                                       ,,***.             .,,,,    ******                       ,,,,*                           # noqa: E501
                                       ,****                       *****,                       ,,**                            # noqa: E501
                                       .***                       ,*****                        , ,                             # noqa: E501
                                                                  *****
                                                                 ,****


                           ***** *      **        ***** **          *******            *****  *                       ***       # noqa: E501
                        ******  *    *****     ******  **** *     *       ***       ******  *                          ***      # noqa: E501
                       **   *  *       *****  **   *  * ****     *         **      **   *  *                            **      # noqa: E501
                      *    *  **       * **  *    *  *   **      **        *      *    *  *                             **      # noqa: E501
                          *  ***      *          *  *             ***                 *  *                              **      # noqa: E501
                         **   **      *         ** **            ** ***              ** **         ****       ****      **      # noqa: E501
                         **   **      *         ** **             *** ***            ** **        * ***  *   * ***  *   **      # noqa: E501
                         **   **     *          ** ******           *** ***        **** **       *   ****   *   ****    **      # noqa: E501
                         **   **     *          ** *****              *** ***     * *** **      **         **    **     **      # noqa: E501
                         **   **     *          ** **                   ** ***       ** **      **         **    **     **      # noqa: E501
                          **  **    *           *  **                    ** **  **   ** **      **         **    **     **      # noqa: E501
                           ** *     *              *                      * *  ***   *  *       **         **    **     **      # noqa: E501
                            ***     *          ****         *   ***        *    ***    *        ***     *  **    **     **      # noqa: E501
                             *******          *  ***********   *  *********      ******          *******    ***** **    *** *   # noqa: E501
                               ***           *     ******     *     *****          ***            *****      ***   **    ***    # noqa: E501
                                             *                *
                                              **               **

logo courtesy Twai (http://www.twitter.com/_twai)
"""


def WhatDoesTheFoxSay():
    print("https://www.youtube.com/watch?v=jofNR_WkoCE")



from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
import warnings as w
from scipy.optimize import root_scalar


class water(model_classes.Model):
    """
    Implementation of the Moore et al. (1998) H2O solubility model for magmas up to 3,000 bars.
    """

    def __init__(self):
        """
        Initialize the model.
        """
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_HB_h2o())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', 3000.0, calibration_checks.crf_LessThan, 'bar',
                'Moore et al. (1998) water',
                fail_msg=calibration_checks.crmsg_LessThan_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'temperature', [700.0, 1200], calibration_checks.crf_Between, 'oC',
                'Moore et al. (1998) water',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])

    def calculate_dissolved_volatiles(self, sample, pressure, temperature, X_fluid=1.0, **kwargs):
        """
        Parameters
        ----------
        sample:     Sample class
            Magma major element chemistry.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 1.0. Mole fraction of H2O in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated dissolved H2O concentration in wt%.
        """

        _sample = sample.change_composition({'H2O': 0.0, 'CO2': 0.0}, inplace=False)

        fH2O = self.fugacity_model.fugacity(pressure=pressure, temperature=temperature,
                                            X_fluid=X_fluid, **kwargs)
        aParam = 2565.0
        bParam_Al2O3 = -1.997
        bParam_FeOt = -0.9275
        bParam_Na2O = 2.736
        cParam = 1.171
        dParam = -14.21

        temperatureK = temperature + 273.15

        sample_molfrac = _sample.get_composition(units='mol_oxides')
        FeOtot = sample_molfrac['FeO'] + sample_molfrac['Fe2O3']*0.8998

        b_x_sum = ((bParam_Al2O3 * sample_molfrac['Al2O3']) + (bParam_FeOt * FeOtot) +
                   (bParam_Na2O * sample_molfrac['Na2O']))
        two_ln_XH2Omelt = ((aParam / temperatureK) + b_x_sum * (pressure/temperatureK) +
                           cParam * np.log(fH2O) + dParam)
        ln_XH2Omelt = two_ln_XH2Omelt / 2.0
        XH2Omelt = np.exp(ln_XH2Omelt)
        sample_molfrac['H2O'] = XH2Omelt

        # Normalize mol fractions to sum to 1, while preserving XH2O
        sample_molfrac = dict(sample_molfrac)
        for key, value in sample_molfrac.items():
            if key != 'H2O':
                sample_molfrac.update({key: value/((1/(1-sample_molfrac['H2O'])))})

        _sample.change_composition(sample_molfrac, units='mol_oxides')

        return _sample.get_composition('H2O')

    def calculate_equilibrium_fluid_comp(self, sample, pressure, temperature, **kwargs):
        """
        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        pressure float
            Pressure in bars.

        temperature float
            Temperature in degrees C.

        Returns
        -------
        float
            Calculated equilibrium fluid concentration in XH2Ofluid mole fraction.
        """

        sample_anhy = sample.change_composition({'H2O': 0.0, 'CO2': 0.0}, inplace=False)

        aParam = 2565.0
        bParam_Al2O3 = -1.997
        bParam_FeOt = -0.9275
        bParam_Na2O = 2.736
        cParam = 1.171
        dParam = -14.21

        temperatureK = temperature + 273.15

        sample_molfrac_anhy = sample_anhy.get_composition(units='mol_oxides')
        sample_molfrac_hy = sample.get_composition(units='mol_oxides')
        FeOtot = sample_molfrac_anhy['FeO'] + sample_molfrac_anhy['Fe2O3']*0.8998

        b_x_sum = ((bParam_Al2O3 * sample_molfrac_anhy['Al2O3']) + (bParam_FeOt * FeOtot) +
                   (bParam_Na2O * sample_molfrac_anhy['Na2O']))
        ln_fH2O = ((2 * np.log(sample_molfrac_hy['H2O']) - (aParam/temperatureK) -
                   b_x_sum * (pressure/temperatureK) - dParam) / cParam)
        fH2O = np.exp(ln_fH2O)
        XH2O_fl = fH2O / pressure

        return XH2O_fl

    def calculate_saturation_pressure(self, temperature, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a an H2O-bearing fluid is saturated. Calls the
        scipy.root_scalar routine, which makes repeated called to the
        calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample:     Sample class
            Magma major element composition.

        temperature float
            Temperature in degrees C.

        X_fluid float
            OPTIONAL. Default is 1.0. Mole fraction of H2O in the H2O-CO2 fluid.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        temperatureK = temperature + 273.15
        if temperatureK <= 0.0:
            raise core.InputError("Temperature must be greater than 0K.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('H2O') is False:
            raise core.InputError("sample must contain H2O.")
        if sample.get_composition('H2O') < 0.0:
            raise core.InputError("Dissolved H2O concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure,
                               args=(temperature, sample, X_fluid, kwargs),
                               x0=100.0, x1=2000.0).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return np.real(satP)

    def root_saturation_pressure(self, pressure, temperature, sample, X_fluid, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample         Sample class
            Magma major element composition.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, X_fluid=X_fluid, **kwargs) -
                sample.get_composition('H2O'))
